#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author:  Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer
#
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 07.07.2023
# Version: 1.0.0
#
# Description:
# This module handles image conversion for the orcaz.
#
# Usage:
# The functions in this module can be imported and used in other modules within the orcaz to perform image conversion.
#
# ----------------------------------------------------------------------------------------------------------------------

import contextlib
import io
import os
import re
import unicodedata
import subprocess
from typing import List

import SimpleITK
import dicom2nifti
import pydicom
from rich.progress import Progress
from orcaz.constants import DCM2NIIX_PATH

def clear_ROI1_and_json(input_path: str) -> None:
    """
    Removes ROI1 files from CT convertion of Siemens CT data

    :param input_path: str, Directory to search and delete files from.
    """
    for root, dirs, files in os.walk(input_path):
        for file in files:
            if ("ROI1") in file:
                os.remove(os.path.join(root, file))
            if file.endswith('.json'):
                os.remove(os.path.join(root, file))


def non_nifti_to_nifti(input_path: str, output_directory: str = None) -> None:
    """
    Converts any image format known to ITK to NIFTI

    :param input_path: str, Directory OR filename to convert to nii.gz
    :param output_directory: str, optional output directory to write the image to.
    """

    if not os.path.exists(input_path):
        print(f"Input path {input_path} does not exist.")
        return

    # Processing a directory
    if os.path.isdir(input_path):
        #modality = get_dicom_modality(input_path)
        nifti_dir = dcm2niix(input_path)
        clear_ROI1_and_json(nifti_dir)
        return

    # Processing a file
    if os.path.isfile(input_path):
        # Ignore hidden or already processed files
        _, filename = os.path.split(input_path)
        if filename.startswith('.') or filename.endswith(('.nii.gz', '.nii')):
            return
        else:
            output_image = SimpleITK.ReadImage(input_path)
            output_image_basename = f"{os.path.splitext(filename)[0]}.nii"

    if output_directory is None:
        output_directory = os.path.dirname(input_path)

    output_image_path = os.path.join(output_directory, output_image_basename)
    SimpleITK.WriteImage(output_image, output_image_path)

# this function returns the path of the folder sin the parent directory
def standardize_to_nifti(parent_dir: str) -> bool:
    """
    Converts all images in a parent directory to NIFTI
    """
    # go through the subdirectories
    subjects = os.listdir(parent_dir)
    # get only the directories
    subjects = [subject for subject in subjects if os.path.isdir(os.path.join(parent_dir, subject))]
    with Progress() as progress:
        task = progress.add_task("[white] Processing subjects...", total=len(subjects))
        for subject in subjects:
            subject_path = os.path.join(parent_dir, subject)
            if os.path.isdir(subject_path):
                image_path = os.path.join(parent_dir, subject)
                non_nifti_to_nifti(image_path,os.path.dirname(image_path))
            elif os.path.isfile(os.path.join(parent_dir, subject)):
                image_path = os.path.join(subject_path, subject)
                print(f"Input path {parent_dir} does not contain subdirectories.")
                return []
            else:
                continue
            progress.update(task, advance=1, description=f"[white] Processing {subject}...")
    if len(subjects)!=0:
        return True
    else:
        return False
        

def is_dicom(file_path) -> bool:
    """
    Checks if a given file is a DICOM image.

    :param str file_path: Path to the file.
    :return: True if the file is a DICOM image, False otherwise.
    :rtype: bool
    """
    try:
        _ = pydicom.dcmread(file_path)
        return True
    except:
        return False

def dcm2niix(input_path: str) -> str:
    """
    Converts DICOM files to NIfTI format using dcm2niix.

    Args:
        input_path (str): The path to the input directory.
        output_dir (str): The path to the output directory.

    Returns:
        str: The path to the output directory.

    Raises:
        NiftiConverterError: If there's an error during DICOM to NIFTI conversion.
    """

    output_dir = os.path.dirname(input_path)
    cmd_to_run: List[str] = [DCM2NIIX_PATH, '-z', 'n', '-f', '%f', '-o', output_dir, input_path]

    try:
        subprocess.run(cmd_to_run, capture_output=True, check=True)
    except subprocess.CalledProcessError:
        raise NiftiConverterError(f"Error during DICOM to NIFTI conversion using dcm2niix.")

    return output_dir

class NiftiConverterError(Exception):
    pass

def remove_accents(unicode_filename):
    try:
        unicode_filename = str(unicode_filename).replace(" ", "_")
        cleaned_filename = unicodedata.normalize('NFKD', unicode_filename).encode('ASCII', 'ignore').decode('ASCII')
        cleaned_filename = re.sub(r'[^\w\s-]', '', cleaned_filename.strip().lower())
        cleaned_filename = re.sub(r'[-\s]+', '-', cleaned_filename)
        return cleaned_filename
    except:
        return unicode_filename


def is_dicom_file(filename):
    try:
        pydicom.dcmread(filename)
        return True
    except pydicom.errors.InvalidDicomError:
        return False


def get_dicom_modality(dicom_dir):
    """Create a lookup dictionary from DICOM files.

    Parameters:
    dicom_dir (str): The directory where DICOM files are stored.

    Returns:
    modality:
    """

    # a dictionary to store information from the DICOM files
    # dicom_info = {}

    # loop over the DICOM files
    for filename in os.listdir(dicom_dir):
        full_path = os.path.join(dicom_dir, filename)
        if is_dicom_file(full_path):
            # read the DICOM file
            ds = pydicom.dcmread(full_path)

            # extract the necessary information
            # series_number = ds.SeriesNumber if 'SeriesNumber' in ds else None
            # series_description = ds.SeriesDescription if 'SeriesDescription' in ds else None
            # sequence_name = ds.SequenceName if 'SequenceName' in ds else None
            # protocol_name = ds.ProtocolName if 'ProtocolName' in ds else None
            # series_instance_UID = ds.SeriesInstanceUID if 'SeriesInstanceUID' in ds else None
            if ds.Modality == 'PT':
                modality = 'PET'
            else:
                modality = ds.Modality

            # # anticipate the filename dicom2nifti will produce and store the modality tag with it
            # if series_number is not None:
            #     base_filename = remove_accents(series_number)
            #     if series_description is not None:
            #         anticipated_filename = f"{base_filename}_{remove_accents(series_description)}.nii"
            #     elif sequence_name is not None:
            #         anticipated_filename = f"{base_filename}_{remove_accents(sequence_name)}.nii"
            #     elif protocol_name is not None:
            #         anticipated_filename = f"{base_filename}_{remove_accents(protocol_name)}.nii"
            # else:
            #     anticipated_filename = f"{remove_accents(series_instance_UID)}.nii"

            # dicom_info[anticipated_filename] = modality

    return modality


def rename_nifti_files(nifti_dir):
    """Rename NIfTI files based on a lookup dictionary.

    Parameters:
    nifti_dir (str): The directory where NIfTI files are stored.
    dicom_info (dict): A dictionary where the key is the anticipated filename that dicom2nifti will produce and
                       the value is the modality of the DICOM series.
    """
    quit()
    # loop over the NIfTI files
    for filename in os.listdir(nifti_dir):
        print("for filename in os.listdir(nifti_dir):", filename)
        if filename.endswith('.nii'):
            # create the new filename
            #new_filename = f"{modality}_{filename}"
            new_filename = f"{modality}_img.nii"
            print("Renaming", filename, "to", new_filename)
            # rename the file
            os.rename(os.path.join(nifti_dir, filename), os.path.join(nifti_dir, new_filename))



import argparse
import os

class Options():
    """This class defines options used during both training and prediction"""

    def __init__(self):
        self.initialized = False

    def initialize(self, parser):
        parser.add_argument('-m', '--mode', help='[pred |coreg |train] switch between training, prediction mode and corregistration (which includes prediction)',
                        required=True)
        # model parameters
        parser.add_argument('--netG', type=str, default='resnet', help='[resnet | Unet]')
        parser.add_argument('--netD', type=str, default='PatchGAN', help='[PatchGAN | PixelGAN]')
        parser.add_argument('--ndf', default=128, type=int, help='number of filters in discriminator')
        parser.add_argument('--ngf', default=64, type=int, help='number of filters in generator')
        parser.add_argument('--generatorLR', type=float, default=0.0002, help='learning rate for generator')
        parser.add_argument('--discriminatorLR', type=float, default=0.0002, help='learning rate for discriminator')
        parser.add_argument('--workers', default=32, type=int, help='number of data loading workers')
        parser.add_argument('--lamb', type=float, default=100, help='weight on L1 term in objective')
        parser.add_argument('--checkpoints_dir', type=str, default='./checkpoints', help='experiments are saved here')
        parser.add_argument('--run_dir', type=str, default='./experiment_AC_PET', help='each training outputs are saved here')
        # parser.add_argument('--generatorWeights', type=str, default='./checkpoints/g.pth', help="path to generator weights (to continue training)")
        # parser.add_argument('--discriminatorWeights', type=str, default='./checkpoints/d.pth', help="path to discriminator weights (to continue training)")

        # basic parameters
        parser.add_argument('--direction', type=str, default='image_to_label', help='image_to_label or label_to_image')
        parser.add_argument('--data_path', type=str, default='./data',help='Location of data for training')
        parser.add_argument('--min_pixel', default=1, help='Percentage of minimum non-zero pixels in the cropped label')
        parser.add_argument('--drop_ratio', default=0, help='Probability to drop a cropped area if the label is empty. All empty patches will be dropped for 0 and accept all cropped patches if set to 1')
        parser.add_argument('--increase_factor_data',  default=2, type=int, help='Increase the data number passed in each epoch')
        parser.add_argument('--output', type=str, default='./checkpoints', help='Training checkpoints save directory')
        parser.add_argument('--gpu_id', type=str, default='0', help='gpu id. Use -1 for CPU')
        parser.add_argument('--save_fre', type=int, default=5, help='model weights save frequency')
        parser.add_argument('--val_fre', type=int, default=5, help='execute validation per epoch frequency')

        # dataset parameters
        parser.add_argument('--resample', default=True, help='Decide or not to rescale the images to a new resolution during training')
        parser.add_argument('--new_resolution', default=(3.3, 3.3, 1.645), help='New resolution')
        #parser.add_argument('--min_pixel', default=1, help='Percentage of minimum non-zero pixels in the cropped label')
        #parser.add_argument('--drop_ratio', default=0, help='Probability to drop a cropped area if the label is empty. All empty patches will be dropped for 0 and accept all cropped patches if set to 1')
        parser.add_argument('--batch_size', type=int, default=2, help='batch size')
        parser.add_argument('--patch_size', default=[128, 128, 128], help='Size of the patches extracted from the image')
        parser.add_argument('--img_channel', default=1, type=int, help='Channels of the image')
        parser.add_argument("--stride_inplane", type=int, nargs=1, default=32, help="Stride size in 2D plane (during validation)")
        parser.add_argument("--stride_layer", type=int, nargs=1, default=32, help="Stride size in z direction (during validation)")

        # training parameters
        parser.add_argument('--epoch_count', type=int, default=1, help='the starting epoch count')
        parser.add_argument('--niter', type=int, default=700, help='# of iter at starting learning rate')
        parser.add_argument('--niter_decay', type=int, default=1450, help='# of iter to linearly decay learning rate to zero')
        parser.add_argument('--lr_policy', type=str, default='lambda', help='learning rate policy: lambda|step|plateau|cosine')
        parser.add_argument('--lr_decay_iters', type=int, default=50, help='multiply by a gamma every lr_decay_iters iterations')
        parser.add_argument('--resume', default=0, type=int, help='resume training or not default:0/not')

        # Inference
        # This is just a trick to make the predict script working
        parser.add_argument("-d", "--subject_directory", type=str,
                            help="Subject directory containing the different PET/CT images of the same subject",
                            required=True)


        self.initialized = True
        return parser

    def parse_options(self):
        if not self.initialized:
            parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                            description='ORCA (Optimized Registration through Conditional Adversarial networks) is a ')
            parser = self.initialize(parser)
        opt = parser.parse_args()
        
        # set gpu ids
        if opt.gpu_id != '-1':
            os.environ["CUDA_VISIBLE_DEVICES"] = opt.gpu_id
        return opt








#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# **********************************************************************************************************************
# File: constants.py
# Project: ORCA
# Created: April 11, 2023, 12:25h
# Author: Lalith Kumar Shiyam Sundar
# Email: lalith.shiyamsundar@meduniwien.ac.at
# Institute: Quantitative Imaging and Medical Physics, Medical University of Vienna
# Description: This module contains the constants for the wolfz.
# License: Apache 2.0
# **********************************************************************************************************************

# Importing required libraries

import os
from datetime import datetime
from orcaz import file_utilities

project_root = file_utilities.get_virtual_env_root()
BINARY_PATH = os.path.join(project_root, 'bin')
MODEL_PATH = os.path.join(project_root, 'models')

# Set the paths to the binaries based on the operating system
if file_utilities.get_system()[0] == 'windows':
    GREEDY_PATH = os.path.join(BINARY_PATH, f'beast-binaries-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                               'greedy.exe')
    C3D_PATH = os.path.join(BINARY_PATH, f'beast-binaries-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                            'c3d.exe')
    DCM2NIIX_PATH = os.path.join(BINARY_PATH,
                                 f'beast-binaries-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                                 'dcm2niix.exe')

elif file_utilities.get_system()[0] in ['linux', 'mac']:
    GREEDY_PATH = os.path.join(BINARY_PATH, f'beast-binaries-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                               'greedy')
    C3D_PATH = os.path.join(BINARY_PATH, f'beast-binaries-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                            'c3d')
    DCM2NIIX_PATH = os.path.join(BINARY_PATH, f'beast-binaries-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                                    'dcm2niix')

# COLOR CODES
ANSI_ORANGE = '\033[38;5;208m'
ANSI_GREEN = '\033[38;5;40m'
ANSI_BLUE = '\033[1;94m'
ANSI_RESET = '\033[0m'
ANSI_VIOLET = '\033[38;5;141m'

# SUPPORTED TRACERS (limited patch)
TRACER_FDG = 'FDG'

# FILE NAMES
RESAMPLED_PREFIX = 'resampled_toPET'
ALIGNED_PREFIX = 'ORCA_aligned_'

# DIRECTORY
ORCA_WORKING_FOLDER = 'ORCA-V01'+'-' + datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
TRANSFORMS_FOLDER = 'transforms'
ALIGNED_CT_FOLDER = 'ORCA_aligned_CT'

# MOOSE PARAMETERS
MOOSE_MODEL = "clin_ct_body"
MOOSE_PREFIX = 'CT_Body_'
MOOSE_LABEL_INDEX = {
        1: "Legs",
        2: "Body",
        3: "Head",
        4: "Arms"
    }
ACCELERATOR = 'cuda'


# EXPECTED MODALITIES
MODALITIES = ['PET', 'CT']
ANATOMICAL_MODALITIES = ['CT']
ANATOMICAL_MODALITIES_SUFFIX = ['_CT']
FUNCTIONAL_MODALITIES = ['PT']
FUNCTIONAL_MODALITIES_SUFFIX = ['_PT']

# EXPECTED PET TYPES (ac, non-ac)
PET_TYPE_AC = ['AC']
PET_TYPE_AC_SUFFIX = ['_AC']
PET_TYPE_NAC = ['NAC']
PET_TYPE_NAC_SUFFIX = ['_NAC']

# TRACER TYPES
TRACER_TYPES = {'FDG':'Fluorodeoxyglucose',
                 'PSMA':'Psma',
                   'FACBC':'Fluciclovine',
                     'DOTA':'Dotatate',
                       'AGNOSTIC':'Agnostic'}
                
# HYPERPARAMETERS
MULTI_RESOLUTION_SCHEME = '100x25x10'

# MOOSE PARAMETERS
MASK_FOLDER = 'masks'


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from tqdm import tqdm
import logging
import matplotlib.pyplot as plt
import math
import torch
from torch.autograd import Variable
import os
import torch
from orcaz.train_utils import *
from orcaz.NiftiDataset import *
from orcaz.registration import *
from orcaz import image_processing, constants
import orcaz.NiftiDataset as NiftiDataset
import torch.optim as optim
import torch.nn as nn
from orcaz.generators import build_netG
from orcaz.options import Options
from orcaz import resources
from orcaz import settings
from rich.progress import Progress

def from_numpy_to_itk(image_np, image_itk):
    image_np = np.transpose(image_np, (2, 1, 0))
    image = sitk.GetImageFromArray(image_np)
    image.SetOrigin(image_itk.GetOrigin())
    image.SetDirection(image_itk.GetDirection())
    image.SetSpacing(image_itk.GetSpacing())
    return image

def prepare_batch(image, ijk_patch_indices):
    image_batches = []
    for batch in ijk_patch_indices:
        image_batch = []
        for patch in batch:
            image_patch = image[patch[0]:patch[1],
                                patch[2]:patch[3], patch[4]:patch[5]]
            image_batch.append(image_patch)

        image_batch = np.asarray(image_batch)
        # image_batch = image_batch[:, :, :, :, np.newaxis]
        image_batches.append(image_batch)

    return image_batches


# inference single image
def inference(write_image, model, image_path, result_path, resolution, patch_size_x, patch_size_y, patch_size_z, stride_inplane, stride_layer, batch_size=1, Logger=True):
    #Get device to run the model
    device = resources.check_cuda(True)

    # Check if image requires resizing
    resample = False
    sitk_image = sitk.ReadImage(image_path)
    original_resolution =  sitk_image.GetSpacing()
    original_size = sitk_image.GetSize()
    if resolution != sitk_image.GetSpacing():
        resample = True
        logging.info("Resampling ...")

    # create transformations to image and labels
    transforms1 = [
        NiftiDataset.Resample(resolution, resample)
    ]

    transforms2 = [
        NiftiDataset.Padding((patch_size_x, patch_size_y, patch_size_z))
    ]

    # read image file
    reader = sitk.ImageFileReader()
    reader.SetFileName(image_path)
    image = reader.Execute()

    # normalize the image
    image = Normalization(image)

    castImageFilter = sitk.CastImageFilter()
    castImageFilter.SetOutputPixelType(sitk.sitkFloat32)
    image = castImageFilter.Execute(image)

    # create empty label in pair with transformed image
    label_tfm = sitk.Image(image.GetSize(), sitk.sitkFloat32)
    label_tfm.SetOrigin(image.GetOrigin())
    label_tfm.SetDirection(image.GetDirection())
    label_tfm.SetSpacing(image.GetSpacing())

    sample = {'image': image, 'label': label_tfm}

    for transform in transforms1:
        sample = transform(sample)

    # keeping track on how much padding will be performed before the inference
    image_array = sitk.GetArrayFromImage(sample['image'])
    pad_x = patch_size_x - (patch_size_x - image_array.shape[2])
    pad_y = patch_size_x - (patch_size_y - image_array.shape[1])
    pad_z = patch_size_z - (patch_size_z - image_array.shape[0])

    image_pre_pad = sample['image']

    for transform in transforms2:
        sample = transform(sample)

    image_tfm, label_tfm = sample['image'], sample['label']

    # convert image to numpy array
    image_np = sitk.GetArrayFromImage(image_tfm)
    label_np = sitk.GetArrayFromImage(label_tfm)

    label_np = np.asarray(label_np, np.float32)

    # unify numpy and sitk orientation
    image_np = np.transpose(image_np, (2, 1, 0))
    label_np = np.transpose(label_np, (2, 1, 0))

    # ----------------- Padding the image if the z dimension still is not even ----------------------

    if (image_np.shape[2] % 2) == 0:
        Padding = False
    else:
        image_np = np.pad(image_np, ((0, 0), (0, 0), (0, 1)), 'edge')
        label_np = np.pad(label_np, ((0, 0), (0, 0), (0, 1)), 'edge')
        Padding = True

    # ------------------------------------------------------------------------------------------------

    # if Weighting option is set for stiching the patches
    
    gaussian_weighting = True
    if gaussian_weighting:
        sigma_scale = [.2, .2, .2]
        patch_size = [patch_size_x, patch_size_y, patch_size_z]
        sigmas = [i * sigma_s for i, sigma_s in zip(patch_size, sigma_scale)]
        importance_map = torch.ones(
            patch_size, device=device, dtype=torch.float)
        for i in range(len(patch_size)):
            x = torch.arange(
                start=-(patch_size[i] - 1) / 2.0, end=(patch_size[i] - 1) / 2.0 + 1, dtype=torch.float, device=device
            )
            x = torch.exp(x**2 / (-2 * sigmas[i] ** 2))  # 1D gaussian
            importance_map = importance_map.unsqueeze(
                -1) * x[(None,) * i] if i > 0 else x

        min_non_zero = max(torch.min(importance_map).item(), 1e-3)
        importance_map = torch.clamp_(importance_map.to(
            torch.float), min=min_non_zero).to(torch.float32)

    logging.info("Preparing patches for prediction ...")

    # a weighting matrix will be used for averaging the overlapped region
    weight_np = np.zeros(label_np.shape)

    # prepare image batch indices
    inum = int(
        math.ceil((image_np.shape[0] - patch_size_x) / float(stride_inplane))) + 1
    jnum = int(
        math.ceil((image_np.shape[1] - patch_size_y) / float(stride_inplane))) + 1
    knum = int(
        math.ceil((image_np.shape[2] - patch_size_z) / float(stride_layer))) + 1

    patch_total = 0
    ijk_patch_indices = []
    ijk_patch_indicies_tmp = []

    for i in range(inum):
        for j in range(jnum):
            for k in range(knum):
                if patch_total % batch_size == 0:
                    ijk_patch_indicies_tmp = []

                istart = i * stride_inplane
                if istart + patch_size_x > image_np.shape[0]:  # for last patch
                    istart = image_np.shape[0] - patch_size_x
                iend = istart + patch_size_x

                jstart = j * stride_inplane
                if jstart + patch_size_y > image_np.shape[1]:  # for last patch
                    jstart = image_np.shape[1] - patch_size_y
                jend = jstart + patch_size_y

                kstart = k * stride_layer
                if kstart + patch_size_z > image_np.shape[2]:  # for last patch
                    kstart = image_np.shape[2] - patch_size_z
                kend = kstart + patch_size_z

                ijk_patch_indicies_tmp.append(
                    [istart, iend, jstart, jend, kstart, kend])

                if patch_total % batch_size == 0:
                    ijk_patch_indices.append(ijk_patch_indicies_tmp)

                patch_total += 1

    batches = prepare_batch(image_np, ijk_patch_indices)

    logging.info("Running prediction ...")
    if Logger is True:

        for i in range(len(batches)):
            batch = batches[i]

            batch = (batch - 127.5) / 127.5

            batch = torch.from_numpy(batch[np.newaxis, :, :, :])
            if device=="cuda":
                batch = Variable(batch.cuda())
            else:
                batch = Variable(batch.cpu())


            pred = model(batch)
            pred = pred.squeeze().data.cpu().numpy()

            pred = (pred * 127.5) + 127.5

            istart = ijk_patch_indices[i][0][0]
            iend = ijk_patch_indices[i][0][1]
            jstart = ijk_patch_indices[i][0][2]
            jend = ijk_patch_indices[i][0][3]
            kstart = ijk_patch_indices[i][0][4]
            kend = ijk_patch_indices[i][0][5]
            label_np[istart:iend, jstart:jend, kstart:kend] += pred[:, :, :] * \
                importance_map.cpu().numpy(
            ) if gaussian_weighting else pred[:, :, :]
            weight_np[istart:iend, jstart:jend,
                      kstart:kend] += importance_map.cpu().numpy() if gaussian_weighting else 1
    else:
        with Progress() as progress:
            task = progress.add_task(f"{constants.ANSI_VIOLET}Processing patches...", total=len(batches))
            for i in range(len(batches)):
                batch = batches[i]

                batch = (batch - 127.5) / 127.5

                batch = torch.from_numpy(batch[np.newaxis, :, :, :])
                if device=="cuda":
                    batch = Variable(batch.cuda())
                else:
                    batch = Variable(batch.cpu())

                pred = model(batch)
                pred = pred.squeeze().data.cpu().numpy()

                pred = (pred * 127.5) + 127.5

                istart = ijk_patch_indices[i][0][0]
                iend = ijk_patch_indices[i][0][1]
                jstart = ijk_patch_indices[i][0][2]
                jend = ijk_patch_indices[i][0][3]
                kstart = ijk_patch_indices[i][0][4]
                kend = ijk_patch_indices[i][0][5]
                label_np[istart:iend, jstart:jend, kstart:kend] += pred[:, :, :] * \
                    importance_map.cpu().numpy(
                ) if gaussian_weighting else pred[:, :, :]
                weight_np[istart:iend, jstart:jend,
                        kstart:kend] += importance_map.cpu().numpy() if gaussian_weighting else 1
                progress.update(task, advance=1, description=f"{constants.ANSI_VIOLET}Processing patch {i+1}/{len(batches)}...")
        logging.info("Prediction complete.")

    # eliminate overlapping region using the weighted value
    label_np = (np.float32(label_np) / np.float32(weight_np) + 0.01)

    # removed the 1 pad on z
    if Padding is True:
        label_np = label_np[:, :, 0:(label_np.shape[2]-1)]

    # removed all the padding
    label_np = label_np[:pad_x, :pad_y, :pad_z]

    # convert back to sitk space
    label = from_numpy_to_itk(label_np, image_pre_pad)
    # ---------------------------------------------------------------------------------------------

    # save label
    writer = sitk.ImageFileWriter()

    if resample is True:
        logging.info("Resampling label back to original image space...")
        label = resample_sitk_image_exact_dimension(
            label, new_spacing=original_resolution,new_size=original_size, interpolator='bspline')
    else:
        label = label
    
    logging.info("Output resolution is "+str(label.GetSize()))

    writer.SetFileName(result_path)
    if write_image is True:
        writer.Execute(label)
        logging.info("Save evaluate label at {} success".format(result_path))
        return label
    else:
        return label

def predict(opt: Options, orca_dir: str, ct_dir: str, pt_dir: str) -> None:

    if resources.check_cuda()=="cuda":
        net = build_netG(opt).cuda()  # load the network Unet
    else:
        net = build_netG(opt)

    net.load_state_dict(new_state_dict(settings.model_path))

    pt_file = glob.glob(os.path.join(pt_dir, '*PT*.nii*'))[0]
    _ = inference(write_image=True,
                  model=net,
                  image_path=pt_file, 
                  result_path=os.path.join(ct_dir, 'synth_ct_img.nii'),
                  resolution=opt.new_resolution,
                  patch_size_x=opt.patch_size[0],
                  patch_size_y=opt.patch_size[1],
                  patch_size_z=opt.patch_size[2],
                  stride_inplane=opt.stride_inplane,
                  stride_layer=opt.stride_layer,
                  Logger=False)

# TODO: To be writen seperately in another file, using the registration Functions from the ENHANCE framework
def motion_match(opt):

    import nibabel
    from scipy.ndimage import binary_dilation

    registrator = image_processing.ImageRegistration(opt.pet, constants.MULTI_RESOLUTION_SCHEME)

    # Load image and mask 
    movingimg = Path(opt.ct)
    ct_nib  = nibabel.load(movingimg.absolute())
    mask_img= nibabel.load(Path(opt.mask).absolute()).get_fdata()
    ct_img  = ct_nib.get_fdata()
    header_info = ct_nib.header

    mask_img = binary_dilation(mask_img, iterations=2)
    minimum = np.median(ct_img[~mask_img])
    ct_img[~mask_img]=minimum

    movingimg = movingimg.parent.absolute() / ("masked_"+movingimg.name)
    img = nibabel.Nifti1Image(ct_img, ct_nib.affine,header_info)
    print("Saving to "+ str(movingimg))
    nibabel.save(img, str(movingimg))

    home_dir = movingimg.parent
    print("Home dir "+str(home_dir))
    tmp_dir = home_dir / Path(opt.dout).name
    tmp_dir.mkdir(parents=True, exist_ok=True)

    mask_img_file = tmp_dir / "mask.nii"
    img = nibabel.Nifti1Image(mask_img, ct_nib.affine,header_info)
    print("Saving to "+str(mask_img_file))
    nibabel.save(img, str(mask_img_file))

    #Create and Save a dilated mask
    dil_mask = binary_dilation(mask_img, iterations=8)
    dil_mask = dil_mask.astype(np.float32)
    
    dmask_img = tmp_dir / "dill_mask.nii"
    mimg = nibabel.Nifti1Image(dil_mask, ct_nib.affine,header_info)
    print("Saving to "+str(dmask_img))
    nibabel.save(mimg, str(dmask_img))
    
    # Saving the filetered image
    flt_img = tmp_dir / str("filetered_"+movingimg.name)
    filter(movingimg, str(flt_img))

    if (opt.gpu_id != -1) & torch.cuda.is_available():
        # device = torch.device("cuda:0")
        os.environ["CUDA_VISIBLE_DEVICES"] = opt.gpu_id
        net = build_netG(opt).cuda()  # load the network Unet
    else:
        net = build_netG(opt)

    net.load_state_dict(new_state_dict(opt.weights))


    tmp_dir = flt_img.parent
    synth_ct_img = tmp_dir / "synth_ct_img.nii"

    _ = inference(write_image=True,
                  model=net,
                  image_path=opt.pet,
                  result_path=str(synth_ct_img),
                  resolution=settings.resolution,
                  patch_size_x=opt.patch_size[0],
                  patch_size_y=opt.patch_size[1],
                  patch_size_z=opt.patch_size[2],
                  stride_inplane=opt.stride_inplane,
                  stride_layer=opt.stride_layer,
                  Logger=False)

    affine(fixed_img=str(synth_ct_img),
           moving_img=movingimg,
           registration_sequence="initial",
           directory=str(tmp_dir),
           cost_function="NCC 2x2x2",
           multi_resolution_iterations="100x50x10",
           nb_threads=opt.workers
           )

    deformable(fixed_img=str(synth_ct_img),
               moving_img=str(flt_img),
               registration_sequence="",
               directory=str(tmp_dir),
               cost_function="NCC 2x2x2",
               multi_resolution_iterations="100x50x10",
               nb_threads=opt.workers,
               mask_img=str(dmask_img)
               )
    # Mask directly the warp file; not used since it reduces the overall effecto  the transformation within the body as well.
    # warp_file =  tmp_dir / str("_warp.nii.gz")
    # warp_nib = nibabel.load(str(warp_file))
    # warp_img = warp_nib.get_fdata()
    # mask = np.bool_(mask_img)
    # warp_img[~mask]=0
    # img = nibabel.Nifti1Image(warp_img, warp_nib.affine,header_info)
    # nibabel.save(img,str(warp_file))
                       
    
    #resampled_img = tmp_dir / str("ORCA_Resampled_"+opt.movingimg)
    resampled_img = tmp_dir / ("ORCA_Resampled_"+str(Path(opt.ct).name))
    resample(fixed_img=str(synth_ct_img),
             moving_img=opt.ct,
             registration_sequence="",
             directory=str(tmp_dir),
             registration_type="deformable",
             resampled_img=str(resampled_img)
             )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# **********************************************************************************************************************
# File: display.py
# Project: 
# Created: April 11, 2023, 13:03h
# Author: Lalith Kumar Shiyam Sundar
# Email: lalith.shiyamsundar@meduniwien.ac.at
# Institute: Quantitative Imaging and Medical Physics, Medical University of Vienna
# Description: A module for displaying the messages of the wolfz processing.
# License: Apache 2.0
# **********************************************************************************************************************

# Importing required libraries

import pyfiglet
from orcaz import constants


def logo():
    """
    Display WOLFZ logo
    :return:
    """
    print(' ')
    logo_color_code = constants.ANSI_BLUE
    slogan_color_code = constants.ANSI_BLUE
    result = logo_color_code + pyfiglet.figlet_format("ORCA 0.1.5", font="slant").rstrip() + "\033[0m"
    text = slogan_color_code + " A part of the ENHANCE-PET framework." + "\033[0m"
    print(result)
    print(text)
    print(' ')


def expectations():
    """
    Display the expectations for the program to run
    :return:
    """
    print(' ')
    print(
        constants.ANSI_ORANGE + " Please make sure that the following conditions are met before running the program:" + constants.ANSI_RESET)
    print(' ')
    print(constants.ANSI_ORANGE + " 1. The subject directory should contain more than one imaging session." +
          constants.ANSI_RESET)
    print(' ')
    print(constants.ANSI_ORANGE + " 2. The subject session should exactly contain one PET and one CT image." +
          constants.ANSI_RESET)
    print(' ')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Project: Optimized Registration through Conditional Adversarial networks (ORCA)
# Author: Zacharias Chalampalakis | Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 3.04.2024
# Version: 0.1.0
#
# Description:
# This module performs input validation for the orca. It verifies that the inputs provided by the user are valid
# and meets the required specifications.
#
# Usage:
# The functions in this module can be imported and used in other modules to perform input validation.
#
# ----------------------------------------------------------------------------------------------------------------------

import logging
from pathlib import Path
import os

from orcaz import constants
from rich.console import Console

console = Console()
def validate_orca_compliant_files_for_pred(path: list) -> bool:
    """
    Selects the files have names that are compliant with orca input files.
    :param path: The path with the present files in the subject directory.
    :return: If input is valid return PET type, else a false flag to terminate execution.
    """
    pet_type = None
    # go through each file and see if the files have the appropriate modality prefix and tracer type.
    files = [file for file in os.listdir(path) if file.endswith('.nii') or file.endswith('.nii.gz')]
    pet_prefixes_files = [file for tag in constants.FUNCTIONAL_MODALITIES_SUFFIX for file in files if tag in file]
    if len(pet_prefixes_files) == 1:
        console.print(f" Orca compliant PET found ", style="bold magenta")
        logging.info(f" Orca compliant PET found")
        if len([file for file in pet_prefixes_files for tag in constants.PET_TYPE_AC_SUFFIX if tag in file]) == 1:
            console.print(f" Orca compliant AC-PET type found ", style="bold magenta")
            logging.info(f" Orca compliant AC-PET type found")
            pet_type = constants.PET_TYPE_AC
        elif len([file for file in pet_prefixes_files for tag in constants.PET_TYPE_NAC_SUFFIX if tag in file]) == 1:
            console.print(f" Orca compliant NAC-PET type found ", style="bold magenta")
            logging.info(f" Orca compliant NAC-PET type found")        
            pet_type = constants.PET_TYPE_NAC
        else: 
            console.print(f" Orca compliant PET type not found ", style="bold magenta")
            logging.info(f" Orca compliant PET type not found")
    else:
        console.print(f" Orca compliant PET not found ", style="bold magenta")
        logging.info(f" Orca compliant PET not found")
    return pet_type

def validate_orca_compliant_files_for_coreg(path: list) -> bool:
    """
    Selects the files have names that are compliant with orca input files.
    :param path: The path with the present files in the subject directory.
    :return: If input is valid return PET type, else a false flag to terminate execution.
    """
    pet_type = None
    # go through each file and see if the files have the appropriate modality prefix and tracer type.
    files = [file for file in os.listdir(path) if file.endswith('.nii') or file.endswith('.nii.gz')]
    pet_prefixes_files = [file for tag in constants.FUNCTIONAL_MODALITIES_SUFFIX for file in files if tag in file]
    ct_prefixes_files = [file for tag in constants.ANATOMICAL_MODALITIES_SUFFIX for file in files if tag in file]
    if len(ct_prefixes_files) == 1 and len(pet_prefixes_files) == 1:
        console.print(f" Orca compliant PET and CT found ", style="bold magenta")
        logging.info(f" Orca compliant PET and CT found")
        if len([file for file in pet_prefixes_files for tag in constants.PET_TYPE_AC_SUFFIX if tag in file]) == 1:
            console.print(f" Orca compliant AC-PET type found ", style="bold magenta")
            logging.info(f" Orca compliant AC-PET type found")
            pet_type = constants.PET_TYPE_AC
        elif len([file for file in pet_prefixes_files for tag in constants.PET_TYPE_NAC_SUFFIX if tag in file]) == 1:
            console.print(f" Orca compliant NAC-PET type found ", style="bold magenta")
            logging.info(f" Orca compliant NAC-PET type found")        
            pet_type = constants.PET_TYPE_NAC
        else: 
            console.print(f" Orca compliant PET type not found ", style="bold magenta")
            logging.info(f" Orca compliant PET type not found")
    else:
        console.print(f" Orca compliant PET and CT not found ", style="bold magenta")
        logging.info(f" Orca compliant PET and CT not found")
    return pet_type

def validate_orca_tracer_type(path: list) -> str:
    """
    Selects the files have tracer names that are compliant with orca input tracer types.
    :param path: The path with the present files in the subject directory.
    :return: If input is avlide return the tracer ID, else an invalide flag to terminate execution.
    """
    tracer_type = None
    # go through each file and see if the files have the appropriate modality prefix and tracer type.
    files = [file for file in os.listdir(path) if file.endswith('.nii') or file.endswith('.nii.gz')]
    pet_prefixes_files = [file for tag in constants.FUNCTIONAL_MODALITIES_SUFFIX for file in files if tag in file]
    if len([file for tag in constants.TRACER_TYPES.keys() for file in pet_prefixes_files if tag in file]) == 1:
        tracer_tag = [tag for tag in constants.TRACER_TYPES.keys() if tag in pet_prefixes_files[0]][0]
        console.print(f" Orca compliant PET tracer {tracer_tag}  found ", style="bold magenta")
        logging.info(f" Orca compliant PET tracer {tracer_tag}  found ")
    else:
        console.print(f" Orca compliant PET tracer not found ", style="bold magenta")
        logging.info(f" Orca compliant PET tracer not found")
    return constants.TRACER_TYPES[tracer_tag]

def combine_pet_type_tag(pet_tag, tracer_tag):
    """
    Combines the PET type and tracer tag to form a unique identifier for the ORCA model.
    :param tag: The PET type tag
    :param tracer_tag: The tracer tag
    :return: The combined PET type and tracer tag
    """
    pet_tag = "nac_" if pet_tag == constants.PET_TYPE_NAC else "ac_" if pet_tag == constants.PET_TYPE_AC else ""
    tracer_tag = tracer_tag.lower()
    return (pet_tag + tracer_tag)


import SimpleITK as sitk
import os
import re
import numpy as np
import random
import glob
import scipy.ndimage.interpolation as interpolation
from sklearn.model_selection import train_test_split
import scipy
import torch
import torch.utils.data
from pathlib import Path


# ------- Swithes -------

interpolator_image = sitk.sitkBSpline                 # interpolator image
interpolator_label = sitk.sitkBSpline                  # interpolator label

_interpolator_image = 'bspline'          # interpolator image
_interpolator_label = 'bspline'          # interpolator label

Segmentation = False

# ------------------------------------- Functions ---------------------------------------


def create_list(data_path):
    """
    this function is create the data list and the data is set as follow:
    --data
        --data_1
            image.nii
            target.nii
        --data_2
            image.nii
            target.nii
        ...
    """
    list_all=[]
    data_path = Path(data_path)
    img_paths = list(data_path.rglob("*image*nii*"))
    lbl_paths = list(data_path.rglob("*label*nii*"))

    image_dict = {str(file_name).rsplit('_',1)[0]: file_name for file_name in img_paths}
    label_dict = {str(file_name).rsplit('_',1)[0]: file_name for file_name in lbl_paths}
    for key in image_dict.keys():
        list_all.extend([{'data': str(image_dict[key]), 'label': str(label_dict[key])}])

    return list_all


def resize(img, new_size, interpolator):
    # img = sitk.ReadImage(img)
    dimension = img.GetDimension()

    # Physical image size corresponds to the largest physical size in the training set, or any other arbitrary size.
    reference_physical_size = np.zeros(dimension)

    reference_physical_size[:] = [(sz - 1) * spc if sz * spc > mx else mx for sz, spc, mx in
                                  zip(img.GetSize(), img.GetSpacing(), reference_physical_size)]

    # Create the reference image with a zero origin, identity direction cosine matrix and dimension
    reference_origin = np.zeros(dimension)
    reference_direction = np.identity(dimension).flatten()
    reference_size = new_size
    reference_spacing = [phys_sz / (sz - 1) for sz, phys_sz in zip(reference_size, reference_physical_size)]

    reference_image = sitk.Image(reference_size, img.GetPixelIDValue())
    reference_image.SetOrigin(reference_origin)
    reference_image.SetSpacing(reference_spacing)
    reference_image.SetDirection(reference_direction)

    # Always use the TransformContinuousIndexToPhysicalPoint to compute an indexed point's physical coordinates as
    # this takes into account size, spacing and direction cosines. For the vast majority of images the direction
    # cosines are the identity matrix, but when this isn't the case simply multiplying the central index by the
    # spacing will not yield the correct coordinates resulting in a long debugging session.
    reference_center = np.array(
        reference_image.TransformContinuousIndexToPhysicalPoint(np.array(reference_image.GetSize()) / 2.0))

    # Transform which maps from the reference_image to the current img with the translation mapping the image
    # origins to each other.
    transform = sitk.AffineTransform(dimension)
    transform.SetMatrix(img.GetDirection())
    transform.SetTranslation(np.array(img.GetOrigin()) - reference_origin)
    # Modify the transformation to align the centers of the original and reference image instead of their origins.
    centering_transform = sitk.TranslationTransform(dimension)
    img_center = np.array(img.TransformContinuousIndexToPhysicalPoint(np.array(img.GetSize()) / 2.0))
    centering_transform.SetOffset(np.array(transform.GetInverse().TransformPoint(img_center) - reference_center))
    centered_transform = sitk.Transform(transform)
    centered_transform.AddTransform(centering_transform)
    # Using the linear interpolator as these are intensity images, if there is a need to resample a ground truth
    # segmentation then the segmentation image should be resampled using the NearestNeighbor interpolator so that
    # no new labels are introduced.

    return sitk.Resample(img, reference_image, centered_transform, interpolator, 0.0)

def resample_sitk_image_exact_dimension(sitk_image,new_spacing=None, new_size=None, interpolator=None, fill_value=0):
    _SITK_INTERPOLATOR_DICT = {
        'nearest': sitk.sitkNearestNeighbor,
        'linear': sitk.sitkLinear,
        'gaussian': sitk.sitkGaussian,
        'label_gaussian': sitk.sitkLabelGaussian,
        'bspline': sitk.sitkBSpline,
        'hamming_sinc': sitk.sitkHammingWindowedSinc,
        'cosine_windowed_sinc': sitk.sitkCosineWindowedSinc,
        'welch_windowed_sinc': sitk.sitkWelchWindowedSinc,
        'lanczos_windowed_sinc': sitk.sitkLanczosWindowedSinc
    }

    orig_pixelid = sitk_image.GetPixelIDValue()
    orig_origin = sitk_image.GetOrigin()
    orig_direction = sitk_image.GetDirection()

    assert interpolator in _SITK_INTERPOLATOR_DICT.keys(), \
        '`interpolator` should be one of {}'.format(_SITK_INTERPOLATOR_DICT.keys())

    sitk_interpolator = _SITK_INTERPOLATOR_DICT[interpolator]

    resampled_sitk_image = sitk.Resample(sitk_image,
                                                   new_size,
                                                   sitk.Transform(),
                                                   sitk_interpolator,
                                                   orig_origin,
                                                   new_spacing,
                                                   orig_direction,
                                                   fill_value,
                                                   orig_pixelid)

    return resampled_sitk_image


def resample_sitk_image(sitk_image, spacing=None, interpolator=None, fill_value=0):
    # https://github.com/SimpleITK/SlicerSimpleFilters/blob/master/SimpleFilters/SimpleFilters.py
    _SITK_INTERPOLATOR_DICT = {
        'nearest': sitk.sitkNearestNeighbor,
        'linear': sitk.sitkLinear,
        'gaussian': sitk.sitkGaussian,
        'label_gaussian': sitk.sitkLabelGaussian,
        'bspline': sitk.sitkBSpline,
        'hamming_sinc': sitk.sitkHammingWindowedSinc,
        'cosine_windowed_sinc': sitk.sitkCosineWindowedSinc,
        'welch_windowed_sinc': sitk.sitkWelchWindowedSinc,
        'lanczos_windowed_sinc': sitk.sitkLanczosWindowedSinc
    }

    """Resamples an ITK image to a new grid. If no spacing is given,
    the resampling is done isotropically to the smallest value in the current
    spacing. This is usually the in-plane resolution. If not given, the
    interpolation is derived from the input data type. Binary input
    (e.g., masks) are resampled with nearest neighbors, otherwise linear
    interpolation is chosen.
    Parameters
    ----------
    sitk_image : SimpleITK image or str
      Either a SimpleITK image or a path to a SimpleITK readable file.
    spacing : tuple
      Tuple of integers
    interpolator : str
      Either `nearest`, `linear` or None.
    fill_value : int
    Returns
    -------
    SimpleITK image.
    """

    if isinstance(sitk_image, str):
        sitk_image = sitk.ReadImage(sitk_image)
    num_dim = sitk_image.GetDimension()

    if not interpolator:
        interpolator = 'linear'
        pixelid = sitk_image.GetPixelIDValue()

        if pixelid not in [1, 2, 4]:
            raise NotImplementedError(
                'Set `interpolator` manually, '
                'can only infer for 8-bit unsigned or 16, 32-bit signed integers')
        if pixelid == 1:  # 8-bit unsigned int
            interpolator = 'nearest'

    orig_pixelid = sitk_image.GetPixelIDValue()
    orig_origin = sitk_image.GetOrigin()
    orig_direction = sitk_image.GetDirection()
    orig_spacing = np.array(sitk_image.GetSpacing())
    orig_size = np.array(sitk_image.GetSize(), dtype=np.int32)

    if not spacing:
        min_spacing = orig_spacing.min()
        new_spacing = [min_spacing] * num_dim
    else:
        new_spacing = [float(s) for s in spacing]

    assert interpolator in _SITK_INTERPOLATOR_DICT.keys(), \
        '`interpolator` should be one of {}'.format(_SITK_INTERPOLATOR_DICT.keys())

    sitk_interpolator = _SITK_INTERPOLATOR_DICT[interpolator]

    new_size = orig_size * (orig_spacing / new_spacing)
    new_size = np.ceil(new_size).astype(np.int32)  # Image dimensions are in integers
    new_size = [int(s) for s in new_size]  # SimpleITK expects lists, not ndarrays

    #resample_filter = sitk.Resample()

    resampled_sitk_image = sitk.Resample(sitk_image,
                                                   new_size,
                                                   sitk.Transform(),
                                                   sitk_interpolator,
                                                   orig_origin,
                                                   new_spacing,
                                                   orig_direction,
                                                   fill_value,
                                                   orig_pixelid)

    return resampled_sitk_image


def matrix_from_axis_angle(a):
    ux, uy, uz, theta = a
    c = np.cos(theta)
    s = np.sin(theta)
    ci = 1.0 - c
    R = np.array([[ci * ux * ux + c,
                   ci * ux * uy - uz * s,
                   ci * ux * uz + uy * s],
                  [ci * uy * ux + uz * s,
                   ci * uy * uy + c,
                   ci * uy * uz - ux * s],
                  [ci * uz * ux - uy * s,
                   ci * uz * uy + ux * s,
                   ci * uz * uz + c],
                  ])
    return R


def resample_image(image, transform):
    reference_image = image
    interpolator = interpolator_image
    default_value = 0
    return sitk.Resample(image, reference_image, transform,
                         interpolator, default_value)


def resample_label(image, transform):
    reference_image = image
    interpolator = interpolator_label
    default_value = 0
    return sitk.Resample(image, reference_image, transform, interpolator, default_value)


def get_center(img):
    width, height, depth = img.GetSize()
    return img.TransformIndexToPhysicalPoint((int(np.ceil(width / 2)),
                                              int(np.ceil(height / 2)),
                                              int(np.ceil(depth / 2))))


def rotation3d_image(image, theta_x, theta_y, theta_z):
    """
    This function rotates an image across each of the x, y, z axes by theta_x, theta_y, and theta_z degrees
    respectively
    :param image: An sitk MRI image
    :param theta_x: The amount of degrees the user wants the image rotated around the x axis
    :param theta_y: The amount of degrees the user wants the image rotated around the y axis
    :param theta_z: The amount of degrees the user wants the image rotated around the z axis
    :param show: Boolean, whether or not the user wants to see the result of the rotation
    :return: The rotated image
    """
    theta_x = np.deg2rad(theta_x)
    theta_y = np.deg2rad(theta_y)
    theta_z = np.deg2rad(theta_z)
    euler_transform = sitk.Euler3DTransform(get_center(image), theta_x, theta_y, theta_z, (0, 0, 0))
    image_center = get_center(image)
    euler_transform.SetCenter(image_center)
    euler_transform.SetRotation(theta_x, theta_y, theta_z)
    resampled_image = resample_image(image, euler_transform)
    return resampled_image


def rotation3d_label(image, theta_x, theta_y, theta_z):
   """
   This function rotates an image across each of the x, y, z axes by theta_x, theta_y, and theta_z degrees
   respectively
   :param image: An sitk MRI image
   :param theta_x: The amount of degrees the user wants the image rotated around the x axis
   :param theta_y: The amount of degrees the user wants the image rotated around the y axis
   :param theta_z: The amount of degrees the user wants the image rotated around the z axis
   :param show: Boolean, whether or not the user wants to see the result of the rotation
   :return: The rotated image
   """
   theta_x = np.deg2rad(theta_x)
   theta_y = np.deg2rad(theta_y)
   theta_z = np.deg2rad(theta_z)
   euler_transform = sitk.Euler3DTransform(get_center(image), theta_x, theta_y, theta_z, (0, 0, 0))
   image_center = get_center(image)
   euler_transform.SetCenter(image_center)
   euler_transform.SetRotation(theta_x, theta_y, theta_z)
   resampled_image = resample_label(image, euler_transform)
   return resampled_image


def flipit(image, axes):
    array = np.transpose(sitk.GetArrayFromImage(image), axes=(2, 1, 0))
    spacing = image.GetSpacing()
    direction = image.GetDirection()
    origin = image.GetOrigin()

    if axes == 0:
        array = np.fliplr(array)
    if axes == 1:
        array = np.flipud(array)

    img = sitk.GetImageFromArray(np.transpose(array, axes=(2, 1, 0)))
    img.SetDirection(direction)
    img.SetOrigin(origin)
    img.SetSpacing(spacing)

    return image


def brightness(image):
    array = np.transpose(sitk.GetArrayFromImage(image), axes=(2, 1, 0))
    spacing = image.GetSpacing()
    direction = image.GetDirection()
    origin = image.GetOrigin()

    max = 255
    min = 0

    c = np.random.randint(-20, 20)

    array = array + c

    array[array >= max] = max
    array[array <= min] = min

    img = sitk.GetImageFromArray(np.transpose(array, axes=(2, 1, 0)))
    img.SetDirection(direction)
    img.SetOrigin(origin)
    img.SetSpacing(spacing)

    return img


def contrast(image):
    array = np.transpose(sitk.GetArrayFromImage(image), axes=(2, 1, 0))
    spacing = image.GetSpacing()
    direction = image.GetDirection()
    origin = image.GetOrigin()

    shape = array.shape
    ntotpixel = shape[0] * shape[1] * shape[2]
    IOD = np.sum(array)
    luminanza = int(IOD / ntotpixel)

    c = np.random.randint(-20, 20)

    d = array - luminanza
    dc = d * abs(c) / 100

    if c >= 0:
        J = array + dc
        J[J >= 255] = 255
        J[J <= 0] = 0
    else:
        J = array - dc
        J[J >= 255] = 255
        J[J <= 0] = 0

    img = sitk.GetImageFromArray(np.transpose(J, axes=(2, 1, 0)))
    img.SetDirection(direction)
    img.SetOrigin(origin)
    img.SetSpacing(spacing)

    return img


def translateit(image, offset, isseg=False):
    order = 0 if isseg == True else 5

    array = np.transpose(sitk.GetArrayFromImage(image), axes=(2, 1, 0))
    spacing = image.GetSpacing()
    direction = image.GetDirection()
    origin = image.GetOrigin()

    array = scipy.ndimage.interpolation.shift(array, (int(offset[0]), int(offset[1]), 0), order=order)

    img = sitk.GetImageFromArray(np.transpose(array, axes=(2, 1, 0)))
    img.SetDirection(direction)
    img.SetOrigin(origin)
    img.SetSpacing(spacing)

    return img


def imadjust(image,gamma=np.random.uniform(1, 2)):

    array = np.transpose(sitk.GetArrayFromImage(image), axes=(2, 1, 0))
    spacing = image.GetSpacing()
    direction = image.GetDirection()
    origin = image.GetOrigin()

    array = (((array - array.min()) / (array.max() - array.min())) ** gamma) * (255 - 0) + 0

    img = sitk.GetImageFromArray(np.transpose(array, axes=(2, 1, 0)))
    img.SetDirection(direction)
    img.SetOrigin(origin)
    img.SetSpacing(spacing)

    return img

# --------------------------------------------------------------------------------------


class NifitDataSet(torch.utils.data.Dataset):

    def __init__(self, data_list,
                 direction='image_to_label',
                 transforms=None,
                 train=False,
                 test=False,):

        # Init membership variables
        self.data_list = data_list
        self.direction = direction
        self.transforms = transforms
        self.train = train
        self.test = test
        self.bit = sitk.sitkFloat32

        """
        the dataset class receive a list that contain the data item, and each item
        is a dict with two item include data path and label path. as follow:
        data_list = [
        {
        "data":　data_path_1,
        "label": label_path_1,
        },
        {
        "data": data_path_2,
        "label": label_path_2,
        }
        ...
        ]
        """

    def read_image(self, path):
        reader = sitk.ImageFileReader()
        reader.SetFileName(path)
        image = reader.Execute()
        return image

    def __getitem__(self, item):

        data_dict = self.data_list[item]
        data_path = data_dict["data"]
        label_path = data_dict["label"]


        if self.direction == 'image_to_label':

            data_path = data_path
            label_path = label_path

        else:

            label_path = data_path
            data_path = label_path

        # read image and label
        image = self.read_image(data_path)

        image = Normalization(image)  # set intensity 0-255

        # cast image and label
        castImageFilter = sitk.CastImageFilter()
        castImageFilter.SetOutputPixelType(self.bit)
        image = castImageFilter.Execute(image)

        if self.train:
            label = self.read_image(label_path)
            if Segmentation is False:
                label = Normalization_CT(label)  # set intensity 0-255
            castImageFilter.SetOutputPixelType(self.bit)
            label = castImageFilter.Execute(label)

        elif self.test:
            label = self.read_image(label_path)
            if Segmentation is False:
                label = Normalization_CT(label)  # set intensity 0-255
            castImageFilter.SetOutputPixelType(self.bit)
            label = castImageFilter.Execute(label)

        else:
            label = sitk.Image(image.GetSize(), self.bit)
            label.SetOrigin(image.GetOrigin())
            label.SetSpacing(image.GetSpacing())

        sample = {'image': image, 'label': label}

        if self.transforms:  # apply the transforms to image and label (normalization, resampling, patches)
            for transform in self.transforms:
                sample = transform(sample)

        # convert sample to tf tensors
        image_np = abs(sitk.GetArrayFromImage(sample['image']))
        label_np = abs(sitk.GetArrayFromImage(sample['label']))

        if Segmentation is True:
            label_np = abs(np.around(label_np))

        # to unify matrix dimension order between SimpleITK([x,y,z]) and numpy([z,y,x])  (actually it´s the contrary)
        image_np = np.transpose(image_np, (2, 1, 0))
        label_np = np.transpose(label_np, (2, 1, 0))

        label_np = (label_np - 127.5) / 127.5
        image_np = (image_np - 127.5) / 127.5

        image_np = image_np[np.newaxis, :, :, :]
        label_np = label_np[np.newaxis, :, :, :]

        return torch.from_numpy(image_np), torch.from_numpy(label_np)  # this is the final output to feed the network

    def __len__(self):
        return len(self.data_list)


def Normalization(image):
    """
    Normalize an image to 0 - 255 (8bits)
    """
    normalizeFilter = sitk.NormalizeImageFilter()
    resacleFilter = sitk.RescaleIntensityImageFilter()
    resacleFilter.SetOutputMaximum(255)
    resacleFilter.SetOutputMinimum(0)

    image = normalizeFilter.Execute(image)  # set mean and std deviation
    image = resacleFilter.Execute(image)  # set intensity 0-255

    return image


def Normalization_CT(image):
    """
    Normalize an image to 0 - 255 (8bits)
    """

    ct_array = sitk.GetArrayFromImage(image)

    upper = +3072
    lower = -1024

    ct_array[ct_array > upper] = upper
    ct_array[ct_array < lower] = lower

    new_ct = sitk.GetImageFromArray(ct_array)
    new_ct.SetDirection(image.GetDirection())
    new_ct.SetOrigin(image.GetOrigin())
    new_ct.SetSpacing(image.GetSpacing())

    normalizeFilter = sitk.NormalizeImageFilter()
    resacleFilter = sitk.RescaleIntensityImageFilter()
    resacleFilter.SetOutputMaximum(255)
    resacleFilter.SetOutputMinimum(0)

    new_ct = normalizeFilter.Execute(new_ct)  # set mean and std deviation
    new_ct = resacleFilter.Execute(new_ct)  # set intensity 0-255

    return new_ct


class StatisticalNormalization(object):
    """
    Normalize an image by mapping intensity with intensity distribution
    """

    def __init__(self, sigma):
        self.name = 'StatisticalNormalization'
        assert isinstance(sigma, float)
        self.sigma = sigma

    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        statisticsFilter = sitk.StatisticsImageFilter()
        statisticsFilter.Execute(image)

        intensityWindowingFilter = sitk.IntensityWindowingImageFilter()
        intensityWindowingFilter.SetOutputMaximum(255)
        intensityWindowingFilter.SetOutputMinimum(0)
        intensityWindowingFilter.SetWindowMaximum(
            statisticsFilter.GetMean() + self.sigma * statisticsFilter.GetSigma());
        intensityWindowingFilter.SetWindowMinimum(
            statisticsFilter.GetMean() - self.sigma * statisticsFilter.GetSigma());

        image = intensityWindowingFilter.Execute(image)

        return {'image': image, 'label': label}


class ManualNormalization(object):
    """
    Normalize an image by mapping intensity with given max and min window level
    """

    def __init__(self, windowMin, windowMax):
        self.name = 'ManualNormalization'
        assert isinstance(windowMax, (int, float))
        assert isinstance(windowMin, (int, float))
        self.windowMax = windowMax
        self.windowMin = windowMin

    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        intensityWindowingFilter = sitk.IntensityWindowingImageFilter()
        intensityWindowingFilter.SetOutputMaximum(255)
        intensityWindowingFilter.SetOutputMinimum(0)
        intensityWindowingFilter.SetWindowMaximum(self.windowMax);
        intensityWindowingFilter.SetWindowMinimum(self.windowMin);

        image = intensityWindowingFilter.Execute(image)

        return {'image': image, 'label': label}


class LaplacianRecursive(object):
    """
    Laplacian recursive image filter
    """

    def __init__(self, sigma):
        self.name = 'Laplacianrecursiveimagefilter'
        assert isinstance(sigma, (int, float))
        self.sigma = sigma


    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        filter = sitk.LaplacianRecursiveGaussianImageFilter()

        filter.SetSigma(1.5)

        image = filter.Execute(image)
        # label = filter.Execute(label)  # comment because is segmentation

        return {'image': image, 'label': label}


class Reorient(object):
    """
    (Beta) Function to orient image in specific axes order
    The elements of the order array must be an permutation of the numbers from 0 to 2.
    """

    def __init__(self, order):
        self.name = 'Reoreient'
        assert isinstance(order, (int, tuple))
        assert len(order) == 3
        self.order = order

    def __call__(self, sample):
        reorientFilter = sitk.PermuteAxesImageFilter()
        reorientFilter.SetOrder(self.order)
        image = reorientFilter.Execute(sample['image'])
        label = reorientFilter.Execute(sample['label'])

        return {'image': image, 'label': label}


class Invert(object):
    """
    Invert the image intensity from 0-255
    """

    def __init__(self):
        self.name = 'Invert'

    def __call__(self, sample):
        invertFilter = sitk.InvertIntensityImageFilter()
        image = invertFilter.Execute(sample['image'], 255)
        label = sample['label']

        return {'image': image, 'label': label}


class Registration(object):

    def __init__(self):
        self.name = 'SurfaceBasedRegistration'

    def __call__(self, sample):
        image, image_sobel, label, label_sobel,  = sample['image'], sample['image'], sample['label'], sample['label']

        Gaus = sitk.GradientMagnitudeRecursiveGaussianImageFilter()
        image_sobel = Gaus.Execute(image_sobel)
        label_sobel = Gaus.Execute(label_sobel)

        fixed_image = label_sobel
        moving_image = image_sobel

        initial_transform = sitk.CenteredTransformInitializer(fixed_image,
                                                              moving_image,
                                                              sitk.Euler3DTransform(),
                                                              sitk.CenteredTransformInitializerFilter.MOMENTS)

        registration_method = sitk.ImageRegistrationMethod()
        # Similarity metric settings.
        registration_method.SetMetricAsMattesMutualInformation(numberOfHistogramBins=50)
        registration_method.SetMetricSamplingStrategy(registration_method.RANDOM)
        registration_method.SetMetricSamplingPercentage(0.1)
        registration_method.SetInterpolator(sitk.sitkLinear)
        # Optimizer settings.
        registration_method.SetOptimizerAsGradientDescent(learningRate=1.0, numberOfIterations=100,
                                                          convergenceMinimumValue=1e-6, convergenceWindowSize=10)
        registration_method.SetOptimizerScalesFromPhysicalShift()

        # Setup for the multi-resolution framework.
        registration_method.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2, 1])
        registration_method.SetSmoothingSigmasPerLevel(smoothingSigmas=[2, 1, 0])
        registration_method.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

        # Don't optimize in-place, we would possibly like to run this cell multiple times.
        registration_method.SetInitialTransform(initial_transform, inPlace=False)

        final_transform = registration_method.Execute(sitk.Cast(fixed_image, sitk.sitkFloat32),
                                                      sitk.Cast(moving_image, sitk.sitkFloat32))

        image = sitk.Resample(image, fixed_image, final_transform, sitk.sitkLinear, 0.0,
                                         moving_image.GetPixelID())

        return {'image': image, 'label': label}


class Resample(object):
    """
    Resample the volume in a sample to a given voxel size

      Args:
          voxel_size (float or tuple): Desired output size.
          If float, output volume is isotropic.
          If tuple, output voxel size is matched with voxel size
          Currently only support linear interpolation method
    """

    def __init__(self, new_resolution, check):
        self.name = 'Resample'
        self.new_resolution = new_resolution
        self.check = check
        # else:
        #     # assert len(new_resolution) == 3
        #     self.new_resolution = new_resolution
        #     self.check = check

    def __call__(self, sample):
        image, label = sample['image'], sample['label']

        new_resolution = self.new_resolution
        check = self.check

        if check is True:
            image = resample_sitk_image(image, spacing=new_resolution, interpolator=_interpolator_image)
            label = resample_sitk_image(label, spacing=new_resolution, interpolator=_interpolator_label)
            #image = resample.Execute(image)
            #label = resample.Execute(label)

            return {'image': image, 'label': label}

        if check is False:
            return {'image': image, 'label': label}

class Padding(object):
    """
    Add padding to the image if size is smaller than patch size

      Args:
          output_size (tuple or int): Desired output size. If int, a cubic volume is formed
      """

    def __init__(self, output_size):
        self.name = 'Padding'

        assert isinstance(output_size, (int, tuple))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size, output_size)
        else:
            assert len(output_size) == 3
            self.output_size = output_size

        assert all(i > 0 for i in list(self.output_size))

    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        size_old = image.GetSize()

        if (size_old[0] >= self.output_size[0]) and (size_old[1] >= self.output_size[1]) and (
                size_old[2] >= self.output_size[2]):
            return sample
        else:
            output_size = self.output_size
            output_size = list(output_size)
            if size_old[0] > self.output_size[0]:
                output_size[0] = size_old[0]
            if size_old[1] > self.output_size[1]:
                output_size[1] = size_old[1]
            if size_old[2] > self.output_size[2]:
                output_size[2] = size_old[2]

            output_size = tuple(output_size)

            resampler = sitk.ResampleImageFilter()
            resampler.SetOutputSpacing(image.GetSpacing())
            resampler.SetSize(output_size)

            # resample on image
            resampler.SetInterpolator(sitk.sitkBSpline)
            resampler.SetOutputOrigin(image.GetOrigin())
            resampler.SetOutputDirection(image.GetDirection())
            image = resampler.Execute(image)

            # resample on label
            resampler.SetInterpolator(sitk.sitkBSpline)
            resampler.SetOutputOrigin(label.GetOrigin())
            resampler.SetOutputDirection(label.GetDirection())

            label = resampler.Execute(label)

            return {'image': image, 'label': label}


class Adapt_eq_histogram(object):
    """
    (Beta) Function to orient image in specific axes order
    The elements of the order array must be an permutation of the numbers from 0 to 2.
    """

    def __init__(self):
        self.name = 'Adapt_eq_histogram'

    def __call__(self, sample):

        adapt = sitk.AdaptiveHistogramEqualizationImageFilter()
        adapt.SetAlpha(0.7)
        adapt.SetBeta(0.8)
        image = adapt.Execute(sample['image'])  # set mean and std deviation

        resacleFilter = sitk.RescaleIntensityImageFilter()
        resacleFilter.SetOutputMaximum(255)
        resacleFilter.SetOutputMinimum(0)
        image = resacleFilter.Execute(image)  # set mean and std deviation

        label = sample['label']

        return {'image': image, 'label': label}


class CropBackground(object):
    """
    Crop the background of the images. Center is fixed in the centroid of the skull
    It crops the images in the xy plane, no cropping is applied to the z direction
    """

    def __init__(self, output_size):
        self.name = 'CropBackground'

        assert isinstance(output_size, (int, tuple))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size, output_size)
        else:
            assert len(output_size) == 3
            self.output_size = output_size

        assert all(i > 0 for i in list(self.output_size))

    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        size_new = self.output_size

        threshold = sitk.BinaryThresholdImageFilter()
        threshold.SetLowerThreshold(1)
        threshold.SetUpperThreshold(255)
        threshold.SetInsideValue(1)
        threshold.SetOutsideValue(0)

        roiFilter = sitk.RegionOfInterestImageFilter()
        roiFilter.SetSize([size_new[0], size_new[1], size_new[2]])

        # label_mask = threshold.Execute(label)
        # label_mask = sitk.GetArrayFromImage(label_mask)
        # label_mask = np.transpose(label_mask, (2, 1, 0))

        image_mask = threshold.Execute(image)
        image_mask = sitk.GetArrayFromImage(image_mask)
        image_mask = np.transpose(image_mask, (2, 1, 0))

        centroid = scipy.ndimage.measurements.center_of_mass(image_mask)

        x_centroid = np.int32(centroid[0])
        y_centroid = np.int32(centroid[1])

        roiFilter.SetIndex([int(x_centroid-(size_new[0])/2), int(y_centroid-(size_new[1])/2), 0])

        label_crop = roiFilter.Execute(label)
        image_crop = roiFilter.Execute(image)

        return {'image': image_crop, 'label': label_crop}


class RandomCrop(object):
    """
    Crop randomly the image in a sample. This is usually used for data augmentation.
      Drop ratio is implemented for randomly dropout crops with empty label. (Default to be 0.2)
      This transformation only applicable in train mode

    Args:
      output_size (tuple or int): Desired output size. If int, cubic crop is made.
    """

    def __init__(self, output_size, drop_ratio=0.1, min_pixel=1):
        self.name = 'Random Crop'

        assert isinstance(output_size, (int, tuple))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size, output_size)
        else:
            assert len(output_size) == 3
            self.output_size = output_size

        assert isinstance(drop_ratio, (int, float))
        if drop_ratio >= 0 and drop_ratio <= 1:
            self.drop_ratio = drop_ratio
        else:
            raise RuntimeError('Drop ratio should be between 0 and 1')

        assert isinstance(min_pixel, int)
        if min_pixel >= 0:
            self.min_pixel = min_pixel
        else:
            raise RuntimeError('Min label pixel count should be integer larger than 0')

    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        size_old = image.GetSize()
        size_new = self.output_size

        contain_label = False

        roiFilter = sitk.RegionOfInterestImageFilter()
        roiFilter.SetSize([size_new[0], size_new[1], size_new[2]])

        # statFilter = sitk.StatisticsImageFilter()     # not useful
        # statFilter.Execute(label)
        # print(statFilter.GetMaximum(), statFilter.GetSum())

        while not contain_label:
            # get the start crop coordinate in ijk
            if size_old[0] <= size_new[0]:
                start_i = 0
            else:
                start_i = np.random.randint(0, size_old[0] - size_new[0])

            if size_old[1] <= size_new[1]:
                start_j = 0
            else:
                start_j = np.random.randint(0, size_old[1] - size_new[1])

            if size_old[2] <= size_new[2]:
                start_k = 0
            else:
                start_k = np.random.randint(0, size_old[2] - size_new[2])

            roiFilter.SetIndex([start_i, start_j, start_k])

            if Segmentation is False:
                # threshold label into only ones and zero
                threshold = sitk.BinaryThresholdImageFilter()
                threshold.SetLowerThreshold(1)
                threshold.SetUpperThreshold(255)
                threshold.SetInsideValue(1)
                threshold.SetOutsideValue(0)
                mask = threshold.Execute(label)
                mask_cropped = roiFilter.Execute(mask)
                label_crop = roiFilter.Execute(label)
                statFilter = sitk.StatisticsImageFilter()
                statFilter.Execute(mask_cropped)  # mine for GANs

            if Segmentation is True:
                label_crop = roiFilter.Execute(label)
                statFilter = sitk.StatisticsImageFilter()
                statFilter.Execute(label_crop)

            # will iterate until a sub volume containing label is extracted
            # pixel_count = seg_crop.GetHeight()*seg_crop.GetWidth()*seg_crop.GetDepth()
            # if statFilter.GetSum()/pixel_count<self.min_ratio:
            if statFilter.GetSum() < self.min_pixel:
                contain_label = self.drop(self.drop_ratio)  # has some probabilty to contain patch with empty label
            else:
                contain_label = True

        image_crop = roiFilter.Execute(image)

        return {'image': image_crop, 'label': label_crop}

    def drop(self, probability):
        return random.random() <= probability


class Augmentation(object):
    """
    Application of transforms. This is usually used for data augmentation.
    List of transforms: random noise
    """

    def __init__(self):
        self.name = 'Augmentation'

    def __call__(self, sample):

        #choice = np.random.choice([0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        choice = np.random.choice([0, 1, 3, 5, 12])
        
        # no augmentation
        if choice == 0:  # no augmentation

            image, label = sample['image'], sample['label']
            return {'image': image, 'label': label}

        # Additive Gaussian noise
        if choice == 1:  # Additive Gaussian noise

            mean = np.random.uniform(0, 1)
            std = np.random.uniform(0, 2)
            self.noiseFilter = sitk.AdditiveGaussianNoiseImageFilter()
            self.noiseFilter.SetMean(mean)
            self.noiseFilter.SetStandardDeviation(std)

            image, label = sample['image'], sample['label']
            image = self.noiseFilter.Execute(image)
            if Segmentation is False:
                label = self.noiseFilter.Execute(label)

            return {'image': image, 'label': label}

        # Recursive Gaussian
        if choice == 2:  # Recursive Gaussian

            sigma = np.random.uniform(0, 1.5)
            self.noiseFilter = sitk.RecursiveGaussianImageFilter()
            self.noiseFilter.SetOrder(0)
            self.noiseFilter.SetSigma(sigma)

            image, label = sample['image'], sample['label']
            image = self.noiseFilter.Execute(image)
            if Segmentation is False:
                label = self.noiseFilter.Execute(label)    # comment for segmentation

            return {'image': image, 'label': label}

        # Random rotation x y z
        if choice == 3:  # Random rotation

            theta_x = np.random.randint(-40, 40)
            theta_y = np.random.randint(-40, 40)
            theta_z = np.random.randint(-180, 180)
            image, label = sample['image'], sample['label']

            image = rotation3d_image(image,theta_x,theta_y, theta_z)
            label = rotation3d_label(label,theta_x,theta_y, theta_z)

            return {'image': image, 'label': label}

        # BSpline Deformation
        if choice == 4:  # BSpline Deformation

            randomness = 10

            assert isinstance(randomness, (int, float))
            if randomness > 0:
                self.randomness = randomness
            else:
                raise RuntimeError('Randomness should be non zero values')

            image, label = sample['image'], sample['label']
            spline_order = 3
            domain_physical_dimensions = [image.GetSize()[0] * image.GetSpacing()[0],
                                          image.GetSize()[1] * image.GetSpacing()[1],
                                          image.GetSize()[2] * image.GetSpacing()[2]]

            bspline = sitk.BSplineTransform(3, spline_order)
            bspline.SetTransformDomainOrigin(image.GetOrigin())
            bspline.SetTransformDomainDirection(image.GetDirection())
            bspline.SetTransformDomainPhysicalDimensions(domain_physical_dimensions)
            bspline.SetTransformDomainMeshSize((10, 10, 10))

            # Random displacement of the control points.
            originalControlPointDisplacements = np.random.random(len(bspline.GetParameters())) * self.randomness
            bspline.SetParameters(originalControlPointDisplacements)

            image = sitk.Resample(image, bspline)
            label = sitk.Resample(label, bspline)
            return {'image': image, 'label': label}

        # Random flip
        if choice == 5:  # Random flip

            axes = np.random.choice([0, 1])
            image, label = sample['image'], sample['label']

            image = flipit(image, axes)
            label = flipit(label, axes)

            return {'image': image, 'label': label}

        # Brightness
        if choice == 6:  # Brightness

            image, label = sample['image'], sample['label']

            image = brightness(image)
            if Segmentation is False:
                label = brightness(label)

            return {'image': image, 'label': label}

        # Contrast
        if choice == 7:  # Contrast

            image, label = sample['image'], sample['label']

            image = contrast(image)
            if Segmentation is False:
                label = contrast(label)             # comment for segmentation

            return {'image': image, 'label': label}

        # Translate
        if choice == 8:  # translate

            image, label = sample['image'], sample['label']

            t1 = np.random.randint(-40, 40)
            t2 = np.random.randint(-40, 40)
            offset = [t1, t2]

            image = translateit(image, offset)
            label = translateit(label, offset)

            return {'image': image, 'label': label}

        # Random rotation z
        if choice == 9:  # Random rotation

            theta_x = 0
            theta_y = 0
            theta_z = np.random.randint(-180, 180)
            image, label = sample['image'], sample['label']

            image = rotation3d_image(image, theta_x, theta_y, theta_z)
            label = rotation3d_label(label, theta_x, theta_y, theta_z)

            return {'image': image, 'label': label}

        # Random rotation x
        if choice == 10:  # Random rotation

            theta_x = np.random.randint(-40, 40)
            theta_y = 0
            theta_z = 0
            image, label = sample['image'], sample['label']

            image = rotation3d_image(image, theta_x, theta_y, theta_z)
            label = rotation3d_label(label, theta_x, theta_y, theta_z)

            return {'image': image, 'label': label}

        # Random rotation y
        if choice == 11:  # Random rotation

            theta_x = 0
            theta_y = np.random.randint(-40, 40)
            theta_z = 0
            image, label = sample['image'], sample['label']

            image = rotation3d_image(image, theta_x, theta_y, theta_z)
            label = rotation3d_label(label, theta_x, theta_y, theta_z)

            return {'image': image, 'label': label}

        # histogram gamma
        if choice == 12:
            image, label = sample['image'], sample['label']

            image = imadjust(image)

            return {'image': image, 'label': label}


class ConfidenceCrop(object):
    """
    Crop the image in a sample that is certain distance from individual labels center.
    This is usually used for data augmentation with very small label volumes.
    The distance offset from connected label centroid is model by Gaussian distribution with mean zero and user input sigma (default to be 2.5)
    i.e. If n isolated labels are found, one of the label's centroid will be randomly selected, and the cropping zone will be offset by following scheme:
    s_i = np.random.normal(mu, sigma*crop_size/2), 1000)
    offset_i = random.choice(s_i)
    where i represents axis direction
    A higher sigma value will provide a higher offset

    Args:
      output_size (tuple or int): Desired output size. If int, cubic crop is made.
      sigma (float): Normalized standard deviation value.
    """

    def __init__(self, output_size, sigma=2.5):
        self.name = 'Confidence Crop'

        assert isinstance(output_size, (int, tuple))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size, output_size)
        else:
            assert len(output_size) == 3
            self.output_size = output_size

        assert isinstance(sigma, (float, tuple))
        if isinstance(sigma, float) and sigma >= 0:
            self.sigma = (sigma, sigma, sigma)
        else:
            assert len(sigma) == 3
            self.sigma = sigma

    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        size_new = self.output_size

        # guarantee label type to be integer
        castFilter = sitk.CastImageFilter()
        castFilter.SetOutputPixelType(sitk.sitkUInt8)
        label = castFilter.Execute(label)

        ccFilter = sitk.ConnectedComponentImageFilter()
        labelCC = ccFilter.Execute(label)

        labelShapeFilter = sitk.LabelShapeStatisticsImageFilter()
        labelShapeFilter.Execute(labelCC)

        if labelShapeFilter.GetNumberOfLabels() == 0:
            # handle image without label
            selectedLabel = 0
            centroid = (int(self.output_size[0] / 2), int(self.output_size[1] / 2), int(self.output_size[2] / 2))
        else:
            # randomly select of the label's centroid
            selectedLabel = random.randint(1, labelShapeFilter.GetNumberOfLabels())
            centroid = label.TransformPhysicalPointToIndex(labelShapeFilter.GetCentroid(selectedLabel))

        centroid = list(centroid)

        start = [-1, -1, -1]  # placeholder for start point array
        end = [self.output_size[0] - 1, self.output_size[1] - 1,
               self.output_size[2] - 1]  # placeholder for start point array
        offset = [-1, -1, -1]  # placeholder for start point array
        for i in range(3):
            # edge case
            if centroid[i] < (self.output_size[i] / 2):
                centroid[i] = int(self.output_size[i] / 2)
            elif (image.GetSize()[i] - centroid[i]) < (self.output_size[i] / 2):
                centroid[i] = image.GetSize()[i] - int(self.output_size[i] / 2) - 1

            # get start point
            while ((start[i] < 0) or (end[i] > (image.GetSize()[i] - 1))):
                offset[i] = self.NormalOffset(self.output_size[i], self.sigma[i])
                start[i] = centroid[i] + offset[i] - int(self.output_size[i] / 2)
                end[i] = start[i] + self.output_size[i] - 1

        roiFilter = sitk.RegionOfInterestImageFilter()
        roiFilter.SetSize(self.output_size)
        roiFilter.SetIndex(start)
        croppedImage = roiFilter.Execute(image)
        croppedLabel = roiFilter.Execute(label)

        return {'image': croppedImage, 'label': croppedLabel}

    def NormalOffset(self, size, sigma):
        s = np.random.normal(0, size * sigma / 2, 100)  # 100 sample is good enough
        return int(round(random.choice(s)))


class BSplineDeformation(object):
    """
    Image deformation with a sparse set of control points to control a free form deformation.
    Details can be found here:
    https://simpleitk.github.io/SPIE2018_COURSE/spatial_transformations.pdf
    https://itk.org/Doxygen/html/classitk_1_1BSplineTransform.html

    Args:
      randomness (int,float): BSpline deformation scaling factor, default is 4.
    """

    def __init__(self, randomness=4):
        self.name = 'BSpline Deformation'

        assert isinstance(randomness, (int, float))
        if randomness > 0:
            self.randomness = randomness
        else:
            raise RuntimeError('Randomness should be non zero values')

    def __call__(self, sample):
        image, label = sample['image'], sample['label']
        spline_order = 3
        domain_physical_dimensions = [image.GetSize()[0] * image.GetSpacing()[0],
                                      image.GetSize()[1] * image.GetSpacing()[1],
                                      image.GetSize()[2] * image.GetSpacing()[2]]

        bspline = sitk.BSplineTransform(3, spline_order)
        bspline.SetTransformDomainOrigin(image.GetOrigin())
        bspline.SetTransformDomainDirection(image.GetDirection())
        bspline.SetTransformDomainPhysicalDimensions(domain_physical_dimensions)
        bspline.SetTransformDomainMeshSize((4, 4, 4))

        # Random displacement of the control points.
        originalControlPointDisplacements = np.random.random(len(bspline.GetParameters())) * self.randomness
        bspline.SetParameters(originalControlPointDisplacements)

        image = sitk.Resample(image, bspline)
        label = sitk.Resample(label, bspline)
        return {'image': image, 'label': label}

    def NormalOffset(self, size, sigma):
        s = np.random.normal(0, size * sigma / 2, 100)  # 100 sample is good enough
        return int(round(random.choice(s)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# **********************************************************************************************************************
# File: config.py
# Project: ORCA
# Created: April 11, 2023, 12:25h
# Author: Zacharias Chalampalakis 
# Email: Zacharias.Chalampalakis@meduniwien.ac.at
# Institute: Quantitative Imaging and Medical Physics, Medical University of Vienna
# Description: This module contains the constants for the wolfz.
# License: Apache 2.0
# **********************************************************************************************************************

import os 
from orcaz import constants
from orcaz.file_utilities import create_directory

orca_dir = ""
synth_dir = ""



import subprocess
import os
import time
from orcaz.file_utilities import *

INITIAL = f'initial'
BONES = f'bones'
ORGANS = f'organs'
TISSUE = f'tissue'
RIGID = f'rigid'
AFFINE = f'affine'
DEFORMABLE = f'deformable'


def filter(moving_img: str, filtered_img: str) -> None:
    """
    Performs filtering of the moving image to match filtering used in training datasets.
    """ #c3d ${3} -smooth 2x2x2mm -o ${1}/$moco_dir/filtered_moving_image.nii.gz 
    
    command_string = f"c3d " \
                     f"{moving_img} " \
                     f"-smooth 2x2x2mm " \
                     f"-o " \
                     f"{filtered_img} "
    
    print(command_string)
    subprocess.run(command_string, shell=True, capture_output=subprocess.DEVNULL)

def rigid(fixed_img: str, moving_img: str, registration_sequence: str, directory: str,cost_function: str, multi_resolution_iterations: str, nb_threads: int) -> None:
    """
    Performs rigid registration between a fixed and moving image using the greedy registration toolkit
    """
    # check if there is an initial transform
    initial_transform_command = f"-ia-image-centers "
    initial_transform_path = "None"
    if os.path.exists(os.path.join(directory, f"{INITIAL}.mat")):
        initial_transform_path = os.path.join(directory, f"{INITIAL}.mat")
        initial_transform_command = f"-ia {initial_transform_path} "

    transform_file_path = os.path.join(directory, f"{registration_sequence}_rigid.mat")
    if registration_sequence == INITIAL:
        transform_file_path = os.path.join(directory, f"{INITIAL}.mat")

    print(f'[GREEDY]    Initial transform: {initial_transform_path}, Output transform: {transform_file_path}')
    command_string = f"greedy " \
                     f"-d 3 " \
                     f"-a " \
                     f"{initial_transform_command}" \
                     f"-i {fixed_img} {moving_img} " \
                     f"-dof 6 " \
                     f"-o {transform_file_path} " \
                     f"-n {multi_resolution_iterations} " \
                     f"-m {cost_function} " \
                     f"-threads {nb_threads} " \
    
    subprocess.run(command_string, shell=True, capture_output=subprocess.DEVNULL)


def affine(fixed_img: str, moving_img: str, registration_sequence: str, directory: str, cost_function: str, multi_resolution_iterations: str, nb_threads: int) -> None:
    """
    Performs affine registration between a fixed and moving image using the greedy registration toolkit
    """
    # check if there is an initial transform
    initial_transform_command = f"-ia-image-centers "
    initial_transform_path = "None"
    if os.path.exists(os.path.join(directory, f"{INITIAL}.mat")):
        initial_transform_path = os.path.join(directory, f"{INITIAL}.mat")
        initial_transform_command = f"-ia {initial_transform_path} "

    transform_file_path = os.path.join(directory, f"{registration_sequence}_affine.mat")
    if registration_sequence == INITIAL:
        transform_file_path = os.path.join(directory, f"{INITIAL}.mat")

    print(f'[GREEDY]    Initial transform: {initial_transform_path}, Output transform: {transform_file_path}')
    command_string = f"greedy " \
                     f"-d 3 " \
                     f"-a " \
                     f"{initial_transform_command}" \
                     f"-i {fixed_img} {moving_img} " \
                     f"-dof 12 " \
                     f"-o {transform_file_path} " \
                     f"-n {multi_resolution_iterations} " \
                     f"-m {cost_function} " \
                     f"-threads {nb_threads} "                                      
    print(command_string)
    subprocess.run(command_string, shell=True, capture_output=subprocess.DEVNULL)


def deformable(fixed_img: str, moving_img: str, registration_sequence: str, directory: str, cost_function: str, multi_resolution_iterations: str, nb_threads: int, mask_img: str) -> None:
    """
    Performs deformable registration between a fixed and moving image using the greedy registration toolkit
    """
    # check if there is an initial transform
    initial_transform_command = f""
    initial_transform_path = ""
    if os.path.exists(os.path.join(directory, f"{INITIAL}.mat")):
        initial_transform_path = os.path.join(directory, f"{INITIAL}.mat")
        initial_transform_command = f"-it {initial_transform_path} "

    transform_file_path = os.path.join(directory, f"{registration_sequence}_warp.nii.gz")
    #inverse_transform = os.path.join(directory, f"{registration_sequence}_inverse_warp.nii.gz")

    print(f'[GREEDY]    Initial transform: {initial_transform_path}, Output transform: {transform_file_path}')
    command_string = f"greedy " \
                     f"-d 3 " \
                     f"-m {cost_function} " \
                     f"{initial_transform_command}" \
                     f"-i {fixed_img} {moving_img} " \
                     f"-o {transform_file_path} " \
                     f"-n {multi_resolution_iterations} " \
                     f"-threads {nb_threads} " \
                     f"-mm {mask_img} " 
                    #f"-oinv {inverse_transform} " #
    print(command_string)
    subprocess.run(command_string, shell=True, capture_output=subprocess.DEVNULL)


def registration(fixed_img: str, moving_img: str, directory: str, registration_sequence: str, registration_type: str, multi_resolution_iterations: str) -> None:
    """
    Registers the fixed and the moving image using the greedy registration toolkit based on the user given cost function
    """

    print(f'[GREEDY]    performing {registration_type} registration for {registration_sequence} to align\n'
          f'            {moving_img} to \n'
          f'            {fixed_img}...')
    start = time.time()

    if registration_type == 'rigid':
        rigid(fixed_img, moving_img, registration_sequence, directory,
              cost_function='SSD', multi_resolution_iterations=multi_resolution_iterations)
    elif registration_type == 'affine':
        affine(fixed_img, moving_img, registration_sequence, directory,
               cost_function='SSD', multi_resolution_iterations=multi_resolution_iterations)
    elif registration_type == 'deformable':
        deformable(fixed_img, moving_img, registration_sequence, directory,
                   cost_function='SSD', multi_resolution_iterations=multi_resolution_iterations)
    else:
        exit("Registration type not supported!")

    end = time.time()
    print(f'            ... completed in {(end - start)/60}min')


def resample(fixed_img: str, moving_img: str, directory: str, registration_sequence: str, registration_type: str, resampled_img:str) -> None:
    """
    Resamples a moving image to match the resolution of a fixed image
    """
    initial_transform_command = f""
    initial_transform_path = ""
    if os.path.exists(os.path.join(directory, f"{INITIAL}.mat")):
        initial_transform_path = os.path.join(directory, f"{INITIAL}.mat")


    print(f"[GREEDY]    resampling {moving_img} based on motion correction results at {directory}.")
    print(f"            writing as {resampled_img}")

    cmd_to_run = ""
    if (registration_type == RIGID or registration_type == AFFINE) and registration_sequence == INITIAL:
        if moving_img and resampled_img:
            cmd_to_run = f"greedy " \
                         f"-d 3 " \
                         f"-rf {fixed_img} " \
                         f"-rm {moving_img} {resampled_img} " \
                         f"-r {os.path.join(directory, f'{registration_sequence}.mat')}"

    elif registration_type == RIGID:
        if moving_img and resampled_img:
            cmd_to_run = f"greedy " \
                         f"-d 3 " \
                         f"-rf {fixed_img} " \
                         f"-rm {moving_img} {resampled_img} " \
                         f"-r {os.path.join(directory, f'{registration_sequence}_rigid.mat')}"

    elif registration_type == AFFINE:
        if moving_img and resampled_img:
            cmd_to_run = f"greedy " \
                         f"-d 3 " \
                         f"-rf {fixed_img} " \
                         f"-rm {moving_img} {resampled_img} " \
                         f"-r {os.path.join(directory, f'{registration_sequence}_affine.mat')}"

    elif registration_type == DEFORMABLE:
        if moving_img and resampled_img:
            cmd_to_run = f"greedy " \
                         f"-d 3 " \
                         f"-rf {fixed_img} " \
                         f"-rm {moving_img} {resampled_img} " \
                         f"-r {os.path.join(directory, f'{registration_sequence}_warp.nii.gz')} {initial_transform_path}"
    print(cmd_to_run)
    subprocess.run(cmd_to_run, shell=True, capture_output=subprocess.DEVNULL)

# def reslice_initial(reference_image, moving_image, resclied_image, initial_transform, interpolation_type="NN"):
#     print(f"[OCELOT]                {moving_image} -> {resclied_image}")
#     print(f"[OCELOT]      Reference {reference_image} | Interpolation: {interpolation_type}")
#     command_string = f"greedy " \
#                      f"-d 3 " \
#                      f"-rf {reference_image} " \
#                      f"-ri {interpolation_type} " \
#                      f"-rm {moving_image} {resclied_image} " \
#                      f"-r {initial_transform}"
#     subprocess.run(command_string, shell=True, capture_output=True)


# def reslice_warp(reference_image, moving_image, resclied_image, warp, initial_transform="", interpolation_type="NN"):
#     print(f"[OCELOT]                {moving_image} -> {resclied_image}")
#     print(f"[OCELOT]      Reference {reference_image} | Interpolation: {interpolation_type}")

#     command_string = f"greedy " \
#                      f"-d 3 " \
#                      f"-rf {reference_image} " \
#                      f"-ri {interpolation_type} " \
#                      f"-rm {moving_image} {resclied_image} " \
#                      f"-r {warp} {initial_transform}"
#     subprocess.run(command_string, shell=True, capture_output=True)



import torch
import torch.nn as nn
import numpy as np
import functools
import random
from orcaz.generators import init_weights


lrelu = nn.LeakyReLU(0.2)


class Flatten(nn.Module):
    def forward(self, input):
        return input.view(input.size(0), -1)


def conv_block(ndf,stage):
    module = []
    module += [
        nn.Conv3d(ndf*(2**stage), ndf*(2**stage), kernel_size=4, stride=2, padding=1),
        nn.BatchNorm3d(ndf*(2**stage)),
        nn.Dropout3d(p=0.5),
        nn.LeakyReLU(negative_slope=0.2),
        nn.Conv3d(ndf*(2**stage), ndf*(2**(stage+1)), kernel_size=4, stride=1, padding=1),
        nn.BatchNorm3d(ndf*(2**(stage+1))),
        nn.Dropout3d(p=0.5),
        nn.LeakyReLU(negative_slope=0.2)
    ]
    return module


class PixelDiscriminator(nn.Module):
    def __init__(self, opt):
        super(PixelDiscriminator, self).__init__()
        self.input_nc = 2
        self.ndf = opt.ndf  # number of filters
        self.norm_layer = nn.BatchNorm3d

        if type(self.norm_layer) == functools.partial:
            use_bias = self.norm_layer.func == nn.InstanceNorm3d
        else:
            use_bias = self.norm_layer == nn.InstanceNorm3d

        self.net = [
            nn.Conv3d(self.input_nc,  self.ndf, kernel_size=1, stride=1, padding=0),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(self.ndf,  self.ndf * 2, kernel_size=1, stride=1, padding=0, bias=use_bias),
            self.norm_layer( self.ndf * 2),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d( self.ndf * 2, 1, kernel_size=1, stride=1, padding=0, bias=use_bias)]

        self.net = nn.Sequential(*self.net)

    def forward(self, input):
        return self.net(input)


class NLayerDiscriminator(nn.Module):
    """Defines a PatchGAN discriminator"""

    def __init__(self, opt):
        super(NLayerDiscriminator, self).__init__()
        self.input_nc = 2
        self.ndf = opt.ndf  # number of filters
        self.n_layers = 3
        self.norm_layer = nn.BatchNorm3d

        if type(self.norm_layer) == functools.partial:  # no need to use bias as BatchNorm2d has affine parameters
            use_bias = self.norm_layer.func == nn.BatchNorm3d
        else:
            use_bias = self.norm_layer == nn.BatchNorm3d

        kw = 4
        padw = 1
        sequence = [nn.Conv3d(self.input_nc, self.ndf, kernel_size=kw, stride=2, padding=padw), nn.LeakyReLU(0.2, True)]

        nf_mult = 1

        for n in range(1, self.n_layers):  # gradually increase the number of filters
            nf_mult_prev = nf_mult
            nf_mult = min(2 ** n, 8)
            sequence += [
                nn.Conv3d(self.ndf * nf_mult_prev, self.ndf * nf_mult, kernel_size=kw, stride=2, padding=padw, bias=use_bias),
                self.norm_layer(self.ndf * nf_mult),
                nn.Dropout(0.2),
                nn.LeakyReLU(0.2, True)
            ]

        nf_mult_prev = nf_mult
        nf_mult = min(2 ** self.n_layers, 8)

        sequence += [
            nn.Conv3d(self.ndf * nf_mult_prev, self.ndf * nf_mult, kernel_size=kw, stride=2, padding=1, bias=use_bias),
            self.norm_layer(self.ndf * nf_mult),
            nn.Dropout(0.2),
            nn.LeakyReLU(0.2, True)
        ]

        sequence += [nn.Conv3d(self.ndf * nf_mult, 1, kernel_size=kw, stride=1, padding=1)]  # output 1 channel prediction map

        self.model = nn.Sequential(*sequence)

    def forward(self, img_input):
        """Standard forward."""
        result = self.model(img_input)
        return result


def build_netD(opt):

    if opt.netD == 'PatchGAN':
        discriminator = NLayerDiscriminator(opt)
    elif opt.netD == 'PixelGAN':
        discriminator = PixelDiscriminator(opt)
    else:
        raise NotImplementedError

    init_weights(discriminator, init_type='normal')

    return discriminator


def smooth_positive_labels(y):
    output = y - 0.3 + (np.random.random(y) * 0.5)

    if output >= 1:

        return 1
    else:
        return output

def smooth_negative_labels(y):
    return 0 + np.random.random(y) * 0.3


class GANLoss(nn.Module):
    def __init__(self, use_lsgan=False, target_real_label=float(smooth_positive_labels(1)), target_fake_label=float(smooth_negative_labels(1))):
        super(GANLoss, self).__init__()
        if random.randint(0, 100) >= 7:  # noisy labels
            self.register_buffer('real_label', torch.tensor(target_real_label))
            self.register_buffer('fake_label', torch.tensor(target_fake_label))
        else:
            self.register_buffer('real_label', torch.tensor(target_fake_label))
            self.register_buffer('fake_label', torch.tensor(target_real_label))
        if use_lsgan:
            self.loss = nn.MSELoss()
        else:
            self.loss = nn.BCEWithLogitsLoss()

    def get_target_tensor(self, input, target_is_real):
        if target_is_real:
            target_tensor = self.real_label
        else:
            target_tensor = self.fake_label
        return target_tensor.expand_as(input)

    def __call__(self, input, target_is_real):
        target_tensor = self.get_target_tensor(input, target_is_real)
        # print(target_tensor),
        # print(target_tensor.shape)
        return self.loss(input, target_tensor)



if __name__ == '__main__':
    import torch
    from torch.autograd import Variable
#s    from torchsummaryX import summary

#    from init import Options
    opt = Options().parse()

    torch.cuda.set_device(0)
    discriminator = build_netD(opt)
    net = discriminator.cuda().eval()

    data= Variable(torch.randn(opt.batch_size, opt.img_channel, opt.patch_size[0], opt.patch_size[1], opt.patch_size[2])).cuda()

    data = torch.cat((data, data), 1)
    out = net(data)

    summary(net, data)
    print("out size: {}".format(out.size()))




import SimpleITK as sitk
import matplotlib.pyplot as plt
from orcaz.NiftiDataset import *
import orcaz.NiftiDataset as NiftiDataset
from orcaz.predict_single_image import inference
from numpy.random import randint

def plot_generated_batch(val_list, model, resolution, patch_size_x, patch_size_y, patch_size_z, stride_inplane, 
stride_layer, batch_size, epoch,save_model_weights_path):
    # --------------- reading the images from the validation list --------------

    image = val_list[0]['data']
    label = val_list[0]['label']

    # ------------------------------------------------------------------------------

    result = inference(False, model, image, 'prova.nii', resolution, patch_size_x,
                       patch_size_y, patch_size_z, stride_inplane, stride_layer, batch_size)

    # save
    #writer = sitk.ImageFileWriter()
    # label_directory = 'checkpoints/Epochs_training/epoch_%s' % epoch
    # if not os.path.exists(label_directory):
    #     os.makedirs(label_directory)
    #label_directory = os.path.join(label_directory, 'epoch_prediction.nii.gz')
    #writer.SetFileName(label_directory)
    #writer.Execute(result)

    # --------------------------- Plotting samples during training ---------------------------------

    first_mod, second_mod, predicted = sitk.ReadImage(image), sitk.ReadImage(label), result

    first_mod = sitk.GetArrayFromImage(first_mod)
    second_mod = sitk.GetArrayFromImage(second_mod)
    predicted = sitk.GetArrayFromImage(result)

    _1 = int(randint(0, first_mod.shape[0], 1) - 1)
    _2 = int(randint(0, first_mod.shape[0], 1) - 1)
    _3 = int(randint(0, first_mod.shape[0], 1) - 1)
    _4 = int(randint(0, first_mod.shape[0], 1) - 1)

    fig = plt.figure()
    fig.set_size_inches(12, 12)

    plt.subplot(5, 3, 1), plt.imshow(first_mod[_1], 'gray'), plt.axis('off'), plt.title('PET')
    plt.subplot(5, 3, 2), plt.imshow(predicted[_1], 'gray'), plt.axis('off'), plt.title('GAN sCT')
    plt.subplot(5, 3, 3), plt.imshow(second_mod[_1], 'gray'), plt.axis('off'), plt.title('CT')

    plt.subplot(5, 3, 4), plt.imshow(first_mod[_2], 'gray'), plt.axis('off'), plt.title('PET')
    plt.subplot(5, 3, 5), plt.imshow(predicted[_2], 'gray'), plt.axis('off'), plt.title('GAN sCT')
    plt.subplot(5, 3, 6), plt.imshow(second_mod[_2], 'gray'), plt.axis('off'), plt.title('CT')

    plt.subplot(5, 3, 7), plt.imshow(first_mod[_3], 'gray'), plt.axis('off'), plt.title('PET')
    plt.subplot(5, 3, 8), plt.imshow(predicted[_3], 'gray'), plt.axis('off'), plt.title('GAN sCT')
    plt.subplot(5, 3, 9), plt.imshow(second_mod[_3], 'gray'), plt.axis('off'), plt.title('CT')

    plt.subplot(5, 3, 10), plt.imshow(first_mod[_4], 'gray'), plt.axis('off'), plt.title('PET')
    plt.subplot(5, 3, 11), plt.imshow(predicted[_4], 'gray'), plt.axis('off'), plt.title('GAN sCT')
    plt.subplot(5, 3, 12), plt.imshow(second_mod[_4], 'gray'), plt.axis('off'), plt.title('CT')

    # plt.subplot(5, 3, 13, autoscale_on=True), plt.hexbin((first_mod / first_mod.max()), (second_mod / second_mod.max()),
    #                                                      bins='log', cmap=plt.cm.Blues,gridsize=50), plt.title('NMI')

    plt.subplot(5, 3, 14, autoscale_on=True), plt.hexbin((predicted), (second_mod),
                                                         bins='log', cmap=plt.cm.Blues,gridsize=50), plt.title('NMI')

    # plt.subplot(5, 3, 15, autoscale_on=True), plt.hexbin((second_mod / second_mod.max()), (second_mod / second_mod.max()), bins='log',
    #                                                      cmap=plt.cm.Blues,gridsize=50), plt.title('NMI')

    plt.savefig('%s/Epoch_{}.png'.format(epoch) % save_model_weights_path)
    plt.close()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# **********************************************************************************************************************
# File: orca.py
# Project: Optimized Registration through Conditional Adversarial networks (ORCA)
# Author: Zacharias Chalampalakis | Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 04.07.2023
# Version: 0.1.5
# Email: zacharias.chalampalakis@meduniwien.ac.at, lalith.shiyamsundar@meduniwien.ac.at
# Description: 
# ORCA is a deep learning based image registration framework for the co-registration between CT and PET images.
# License: Apache 2.0
# Usage:
# python orca.py --mode coreg --ct /path/to/ct.nii.gz --pet /path/to/pet.nii.gz --dout /path/to/output_dir
# python orca.py --mode pred --pet /path/to/pet.nii.gz --dout /path/to/output_dir
# python orca.py --mode train --data_path /path/to/data --output /path/to/checkpoints_dir
# **********************************************************************************************************************

# Importing required libraries
import logging
import sys
import emoji
import time

from datetime import datetime
from orcaz import train
from orcaz import predict_single_image
from orcaz import display,constants,file_utilities
from orcaz import download,resources
from orcaz.options import Options
from orcaz import image_conversion, input_validation
from orcaz import settings
from orcaz import image_processing
from pathlib import Path
import os 
import glob


import warnings
warnings.filterwarnings('ignore', category=UserWarning, message='TypedStorage is deprecated')

import resource
rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (4096, rlimit[1]))

logging.basicConfig(format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s', level=logging.INFO,
                    filename=datetime.now().strftime('orca-%H-%M-%d-%m-%Y.log'),
                    filemode='w')
# uncomment the following line to print the logs to the console
#logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))

def main():

    display.logo()
    logging.info("----------------------------------------------------------------------------------------------------")
    logging.info("                                         STARTING ORCA 0.1.0                                        ")
    logging.info("----------------------------------------------------------------------------------------------------")
    logging.info(' ')

    # ----------------------------------
    # INPUT ARGUMENTS and SWITCH BETWEEN MODES
    # ----------------------------------
   
    opt = Options().parse_options()
    # TODO: Add subparsers for the different options.
    if opt.mode =='train':
        logging.info('-'*50)
        logging.info('          Using ORCA in training mode             ')
        logging.info('-'*50)
        train.train(opt)

    elif opt.mode =='pred':
        logging.info('-'*50)
        logging.info('        Using ORCA in prediction mode        ')
        logging.info('-'*50) 

        # ----------------------------------
        # INPUT STANDARDIZATION
        # ----------------------------------
        # calculate elapsed time for the entire procedure below
        start_time = time.time()
        print('')
        print(f'{constants.ANSI_BLUE} {emoji.emojize(":magnifying_glass_tilted_left:")} STANDARDIZING INPUT DATA TO '
            f'NIFTI:{constants.ANSI_RESET}')
        print('')
        logging.info(' ')
        logging.info(' STANDARDIZING INPUT DATA:')
        logging.info(' ')
        download.download_binaries()
        dcm_flag = image_conversion.standardize_to_nifti(opt.subject_directory)
        print(f"{constants.ANSI_GREEN} Standardization complete.{constants.ANSI_RESET}")
        logging.info(" Standardization complete.")
        
        pet_type =  input_validation.validate_orca_compliant_files_for_pred(opt.subject_directory)
        if pet_type == None:
            return 1

        tracer_tag = input_validation.validate_orca_tracer_type(opt.subject_directory)
        if tracer_tag == None:
            return 1
        
        # Select the approriate model type based on the PET type and tracer tag
        model_id = input_validation.combine_pet_type_tag(pet_type, tracer_tag)
        models_path = constants.MODEL_PATH
        file_utilities.create_directory(models_path)
        download.download(item_name=model_id, item_path=constants.MODEL_PATH, item_dict=resources.ORCA_MODELS)
        settings.model_path = os.path.join(constants.MODEL_PATH,resources.ORCA_MODELS[model_id]['directory'],resources.ORCA_MODELS[model_id]['model'])
        settings.resolution = resources.ORCA_MODELS[model_id]['voxel_spacing']

        # ---------------------------------------------------
        # RUNNING PREPROCESSING AND PREDICTION PIPELINE
        # ---------------------------------------------------
        start_time = time.time()
        print('')
        print(f'{constants.ANSI_VIOLET} {emoji.emojize(":rocket:")} RUNNING PREPROCESSING AND PREDICTION PIPELINE:{constants.ANSI_RESET}')
        print('')      

        orca_dir, ct_dir, pt_dir, mask_dir = image_processing.preprocess(opt.subject_directory,dicom_files_flag=dcm_flag, coreg_flag=False)
        print(f'{constants.ANSI_GREEN} {emoji.emojize(":hourglass_done:")} Preprocessing complete.{constants.ANSI_RESET}')
        print(f'{constants.ANSI_VIOLET} {emoji.emojize(":robot:")} Generating syntheticCT.{constants.ANSI_RESET}')
        predict_single_image.predict(opt, orca_dir, orca_dir, pt_dir)


    elif opt.mode=='coreg':
        logging.info('-'*50)
        logging.info('        Using ORCA in co-registration mode        ')
        logging.info('-'*50)

        # ----------------------------------
        # INPUT STANDARDIZATION
        # ----------------------------------
        # calculate elapsed time for the entire procedure below
        start_time = time.time()
        print('')
        print(f'{constants.ANSI_BLUE} {emoji.emojize(":magnifying_glass_tilted_left:")} STANDARDIZING INPUT DATA TO '
            f'NIFTI:{constants.ANSI_RESET}')
        print('')
        logging.info(' ')
        logging.info(' STANDARDIZING INPUT DATA:')
        logging.info(' ')
        download.download_binaries()
        dcm_flag = image_conversion.standardize_to_nifti(opt.subject_directory)
        print(f"{constants.ANSI_GREEN} Standardization complete.{constants.ANSI_RESET}")
        logging.info(" Standardization complete.")
        
        pet_type =  input_validation.validate_orca_compliant_files_for_coreg(opt.subject_directory)
        if pet_type == None:
            return 1

        tracer_tag = input_validation.validate_orca_tracer_type(opt.subject_directory)
        if tracer_tag == None:
            return 1
        
        # Select the approriate model type based on the PET type and tracer tag
        model_id = input_validation.combine_pet_type_tag(pet_type, tracer_tag)
        models_path = constants.MODEL_PATH
        file_utilities.create_directory(models_path)
        download.download(item_name=model_id, item_path=constants.MODEL_PATH, item_dict=resources.ORCA_MODELS)
        settings.model_path = os.path.join(constants.MODEL_PATH,resources.ORCA_MODELS[model_id]['directory'],resources.ORCA_MODELS[model_id]['model'])
        settings.resolution = resources.ORCA_MODELS[model_id]['voxel_spacing']

        # ---------------------------------------------------
        # RUNNING PREPROCESSING AND CO-REGISTRATION PIPELINE
        # ---------------------------------------------------
        start_time = time.time()
        print('')
        print(f'{constants.ANSI_VIOLET} {emoji.emojize(":rocket:")} RUNNING PREPROCESSING AND CO-REGISTRATION PIPELINE:{constants.ANSI_RESET}')
        print('')

        logging.info(' ')
        logging.info(' RUNNING PREPROCESSING AND CO-REGISTRATION PIPELINE:')
        logging.info(' ')
        orca_dir, ct_dir, pt_dir, mask_dir = image_processing.preprocess(opt.subject_directory,dicom_files_flag=dcm_flag, coreg_flag=True)
        print(f'{constants.ANSI_GREEN} {emoji.emojize(":hourglass_done:")} Preprocessing complete.{constants.ANSI_RESET}')
        print(f'{constants.ANSI_VIOLET} {emoji.emojize(":robot:")} Generating syntheticCT.{constants.ANSI_RESET}')
        predict_single_image.predict(opt, orca_dir, ct_dir, pt_dir)
        print(f'{constants.ANSI_GREEN} {emoji.emojize(":hourglass_done:")} synthetic CT generation complete.{constants.ANSI_RESET}')
        print(f'{constants.ANSI_VIOLET} {emoji.emojize(":whale:")} Co-registering CT to synthetic CT.{constants.ANSI_RESET}')
        image_processing.align(orca_dir, ct_dir, pt_dir, mask_dir,opt.workers)
        image_processing.postprocess(orca_dir, ct_dir, mask_dir)
        # Check if input ct is in DICOM format, if so export the co-registered CT to DICOM
        subject_directory = os.path.abspath(opt.subject_directory)
        if dcm_flag and image_conversion.is_dicom(glob.glob(os.path.join(subject_directory, '*_CT/*'))[0]):
            image_processing.export_dicom(orca_dir, opt.subject_directory)
        end_time = time.time()
        elapsed_time = end_time - start_time
        # show elapsed time in minutes and round it to 2 decimal places
        elapsed_time = round(elapsed_time / 60, 2)
        print(f'{constants.ANSI_GREEN} {emoji.emojize(":hourglass_done:")} Co-registration complete.'
            f' Elapsed time: {elapsed_time} minutes! {emoji.emojize(":partying_face:")} \n Co-registered images are stored in'
            f' {constants.ALIGNED_CT_FOLDER}! {constants.ANSI_RESET}')

    else:
        logging.error("*** !Unknown mode option requested, exiting now! ***")
        print(f'{constants.ANSI_ORANGE} {emoji.emojize(":warning:")} Unknown mode option requested, exiting now! {constants.ANSI_RESET}')
        return 1; 

if __name__ == '__main__':
  main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# **********************************************************************************************************************
# File: train.py
# Project: Optimized Registration through Conditional Adversarial networks (ORCA)
# Created: April 11, 2023, 12:01h
# Authors: Zacharias Chalampalakis, Lalith Kumar Shiyam Sundar
# Email: zacharias.chalampalakis@meduniwien.ac.at, lalith.shiyamsundar@meduniwien.ac.at
# Institute: Quantitative Imaging and Medical Physics, Medical University of Vienna
# Description:
# License: Apache 2.0
# **********************************************************************************************************************

# Importing required libraries
import sys
import os
import torch
import torch.optim as optim
import torch.nn as nn
import glob
import logging
from math import log10
import tqdm

from orcaz.generators import *
from orcaz.discriminators import *
from orcaz.train_utils import *
from orcaz.file_utilities import *
from orcaz.train_utils import *
import orcaz.NiftiDataset as NiftiDataset
from orcaz.NiftiDataset import *
from orcaz.plot_generated_batch import plot_generated_batch
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter


def train(opt):

    # TODO: Selectively get device number
    if (opt.gpu_id != -1) & torch.cuda.is_available():
        #device = torch.device("cuda:0")
        os.environ["CUDA_VISIBLE_DEVICES"] = opt.gpu_id
    
    create_directory(opt.checkpoints_dir)
    output_dir = Path(opt.checkpoints_dir)
    save_model_weights_path = output_dir / opt.run_dir

    if save_model_weights_path.exists() and (save_model_weights_path / "checkpoint.pth").exists():
        resume = True
    else:
        resume = False
    save_model_weights_path.mkdir(exist_ok=True)

    # -----  Loading the list of data -----
    train_list = create_list(opt.data_path + "/train")
    val_list = create_list(opt.data_path + "/val")

    for i in range(opt.increase_factor_data):  # augment the data list for training
        train_list.extend(train_list)
        val_list.extend(val_list)

    print('Number of training patches per epoch:', len(train_list))
    print('Number of validation patches per epoch:', len(val_list))
    # -------------------------------------

    ############## TENSORBOARD ########################
    from torch.utils.tensorboard import SummaryWriter
    # default `log_dir` is "runs" - we'll be more specific here
    writer_train = SummaryWriter(str(save_model_weights_path)+"/train/")
    writer_val = SummaryWriter(str(save_model_weights_path)+"/val/")
    ###################################################

    # -----  Transformation and Augmentation process for the data  -----
    min_pixel = int(opt.min_pixel * ((opt.patch_size[0] * opt.patch_size[1] * opt.patch_size[2]) / 100))
    trainTransforms = [
                NiftiDataset.Resample(opt.new_resolution, opt.resample),
                NiftiDataset.Augmentation(),
                NiftiDataset.Padding((opt.patch_size[0], opt.patch_size[1], opt.patch_size[2])),
                NiftiDataset.RandomCrop((opt.patch_size[0], opt.patch_size[1], opt.patch_size[2]), opt.drop_ratio, min_pixel),
                ]

    train_set = NifitDataSet(train_list, direction=opt.direction, transforms=trainTransforms, train=True)    # define the dataset and loader
    train_loader = DataLoader(train_set, batch_size=opt.batch_size, shuffle=True, num_workers=opt.workers)  # Here are then fed to the network with a defined batch size

    valTransforms = [
        NiftiDataset.Resample(opt.new_resolution, opt.resample),
        NiftiDataset.Padding((opt.patch_size[0], opt.patch_size[1], opt.patch_size[2])),
        NiftiDataset.RandomCrop((opt.patch_size[0], opt.patch_size[1], opt.patch_size[2]), opt.drop_ratio, min_pixel),
    ]

    val_set = NifitDataSet(val_list, direction=opt.direction, transforms=valTransforms, test=True)
    val_loader = DataLoader(val_set, batch_size=opt.batch_size, shuffle=False, num_workers=opt.workers)


    # -----  Creating the Generator, discriminator and optimizers/schedulers -----
    generator = build_netG(opt)
    discriminator = build_netD(opt)

    criterionMSE = nn.MSELoss()  # nn.MSELoss()
    criterionGAN = GANLoss()
    criterion_pixelwise = nn.L1Loss()
    # -----  Use Single GPU or Multiple GPUs -----
    if (opt.gpu_id != -1) & torch.cuda.is_available():
        use_gpu = True
        generator.cuda()
        discriminator.cuda()
        criterionGAN.cuda()
        criterion_pixelwise.cuda()

        # if num_gpus > 1:
        #     generator = nn.DataParallel(generator)
        #     discriminator = nn.DataParallel(discriminator)

    optim_generator = optim.Adam(generator.parameters(), betas=(0.5,0.999), lr=opt.generatorLR)
    optim_discriminator = optim.Adam(discriminator.parameters(), betas=(0.5,0.999), lr=opt.discriminatorLR)
    net_g_scheduler = get_scheduler(optim_generator, opt)
    net_d_scheduler = get_scheduler(optim_discriminator, opt)

    if resume:
        print(f"Using checkpoint!")
        checkpoint = torch.load(str(save_model_weights_path / "checkpoint.pth"))
        generator.load_state_dict(checkpoint["state_dict"])
        discriminator.load_state_dict(checkpoint["discriminator"])
        optim_generator.load_state_dict(checkpoint["optimizer_g"])
        optim_discriminator.load_state_dict(checkpoint["optimizer_d"])
        net_g_scheduler.load_state_dict(checkpoint["net_g_scheduler"])
        net_d_scheduler.load_state_dict(checkpoint["net_d_scheduler"])
        epoch_count= checkpoint["epoch"]
    else:
        epoch_count = opt.epoch_count

    # -----  Training Cycle -----
    logging.info('Start training :) ')
    
    for epoch in range(epoch_count, opt.niter + opt.niter_decay):
        mean_generator_total_loss = 0.0
        mean_discriminator_loss = 0.0
        step = 0
        generator.train()

        pbar = tqdm.tqdm(enumerate(train_loader), total=len(train_loader))
        for batch_idx, (data, label) in pbar:

            real_a = data
            real_b = label

            if use_gpu:                              # forward
                real_b = real_b.cuda()
                fake_b = generator(real_a.cuda())    # generate fake data
                real_a = real_a.cuda()
            else:
                fake_b = generator(real_a)

            ######################
            # (1) Update D network
            ######################
            optim_discriminator.zero_grad()

            # train with fake
            fake_ab = torch.cat((real_a, fake_b), 1)
            pred_fake = discriminator.forward(fake_ab.detach())
            loss_d_fake = criterionGAN(pred_fake, False)

            # train with real
            real_ab = torch.cat((real_a, real_b), 1)
            pred_real = discriminator.forward(real_ab)
            loss_d_real = criterionGAN(pred_real, True)

            # Combined D loss
            discriminator_loss = (loss_d_fake + loss_d_real) * 0.5

            mean_discriminator_loss += discriminator_loss
            discriminator_loss.backward()
            optim_discriminator.step()

            ######################
            # (2) Update G network
            ######################

            optim_generator.zero_grad()

            # First, G(A) should fake the discriminator
            fake_ab = torch.cat((real_a, fake_b), 1)
            pred_fake = discriminator.forward(fake_ab)
            loss_g_gan = criterionGAN(pred_fake, True)

            # Second, G(A) = B
            loss_g_l1 = criterion_pixelwise(fake_b, real_b) * opt.lamb

            generator_total_loss = loss_g_gan + loss_g_l1

            mean_generator_total_loss += generator_total_loss
            generator_total_loss.backward()
            optim_generator.step()

            ######### Status and display #########
            pbar.set_postfix(
            {
                "epoch": epoch_count,
                "Discriminator_Loss": f"{discriminator_loss:.6f}",
                "Generator_Loss": f"{generator_total_loss:.6f}",
                # "p_loss": f"{losses['p_loss'].item():.6f}",
                # "g_loss": f"{losses['g_loss'].item():.6f}",
                # "d_loss": f"{losses['d_loss'].item():.6f}",
                # "lr_g": f"{get_lr(optimizer_g):.6f}",
                # "lr_d": f"{get_lr(optimizer_d):.6f}",
            })

        update_learning_rate(net_g_scheduler, optim_generator)
        update_learning_rate(net_d_scheduler, optim_discriminator)

        ##### Logger ######
        mean_generator_total_loss /= len(train_loader)
        mean_discriminator_loss /= len(train_loader)

        if (epoch) % opt.val_fre == 0:
            plot_generated_batch(val_list=val_list, model=generator, resolution=opt.new_resolution,
                                patch_size_x=opt.patch_size[0], patch_size_y=opt.patch_size[1],
                                patch_size_z=opt.patch_size[2], stride_inplane=opt.stride_inplane,
                                stride_layer=opt.stride_layer, batch_size=1,
                                epoch=epoch_count, save_model_weights_path=save_model_weights_path)

        ############## TENSORBOARD ########################
        writer_train.add_scalar('generator_total_loss',
                            mean_generator_total_loss.cpu().item(), epoch)
        writer_train.add_scalar('discriminator_loss',
                            mean_discriminator_loss.cpu().item(), epoch)
        ###################################################
    
        epoch_count += 1
        if epoch % opt.save_fre == 0:
            torch.save(generator.state_dict(), '%s/g_epoch_{}.pth'.format(epoch) % save_model_weights_path)
            torch.save(discriminator.state_dict(), '%s/d_epoch_{}.pth'.format(epoch) % save_model_weights_path)

            # Save checkpoint
            checkpoint = {
                "epoch": epoch_count,
                "state_dict": generator.state_dict(),
                "discriminator": discriminator.state_dict(),
                "optimizer_g": optim_generator.state_dict(),
                "optimizer_d": optim_discriminator.state_dict(),
                "net_g_scheduler": net_g_scheduler.state_dict(),
                "net_d_scheduler": net_d_scheduler.state_dict()
            }
            torch.save(checkpoint, str(save_model_weights_path / "checkpoint.pth"))

        # sys.stdout.write(
        #     '\r[%d/%d] Discriminator_Loss: %.4f Generator_Loss:%.4f \n' % (
        #         epoch_count-1, (opt.niter + opt.niter_decay + 1),
        #         mean_discriminator_loss / len(train_loader),
        #         mean_generator_total_loss / len(train_loader)
        #         ))



import os
import torch
import torch.nn as nn
import numpy as np
from scipy.ndimage import morphology
from collections import OrderedDict
from torch.optim import lr_scheduler
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, Mapping, Hashable


def get_scheduler(optimizer, opt):
    if opt.lr_policy == 'lambda':
        def lambda_rule(epoch):
            lr_l = 1.0 - max(0, epoch + opt.epoch_count - opt.niter) / float(opt.niter_decay + 1)
            return lr_l
        scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda_rule)
    elif opt.lr_policy == 'step':
        scheduler = lr_scheduler.StepLR(optimizer, step_size=opt.lr_decay_iters, gamma=0.1)
    elif opt.lr_policy == 'plateau':
        scheduler = lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.2, threshold=0.01, patience=5)
    elif opt.lr_policy == 'cosine':
        scheduler = lr_scheduler.CosineAnnealingLR(optimizer, T_max=opt.niter, eta_min=0)
    else:
        return NotImplementedError('learning rate policy [%s] is not implemented', opt.lr_policy)
    return scheduler


# update learning rate (called once every epoch)
def update_learning_rate(scheduler, optimizer):
    scheduler.step()
    lr = optimizer.param_groups[0]['lr']
    # print('learning rate = %.7f' % lr)


def Cor_CoeLoss(y_pred, y_target):
    x = y_pred
    y = y_target
    x_var = x - torch.mean(x)
    y_var = y - torch.mean(y)
    r_num = torch.sum(x_var * y_var)
    r_den = torch.sqrt(torch.sum(x_var ** 2)) * torch.sqrt(torch.sum(y_var ** 2))
    r = r_num / r_den

    # return 1 - r  # best are 0
    return 1 - r**2 # abslute constrain


def dice_coeff(seg, gt, ratio=0.5):
    """
    function to calculate the dice score
    """
    seg = seg.flatten()
    gt = gt.flatten()
    seg[seg > ratio] = np.float32(1)
    seg[seg < ratio] = np.float32(0)
    dice = float(2 * (gt * seg).sum())/float(gt.sum() + seg.sum())
    return dice


def check_dir(path):              # if folder does not exist, create it
    if not os.path.exists(path):
        os.mkdir(path)


def new_state_dict(file_name):
    state_dict = torch.load(file_name,weights_only=True)
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        if k[:6] == 'module':
            name = k[7:]
            new_state_dict[name] = v
        else:
            new_state_dict[k] = v
    return new_state_dict
